import numpy as np


def calc_mag(amp, event_loc, station_loc, weight, min=-2, max=8):
    """
    Calculate magnitude from amplitude and distance
    Input:
        amp: amplitude in log10(cm/s)
        event_loc: event location (n, 3)
        station_loc: station location (n, 3)
        weight: weight for each observation (n,)
        min: minimum magnitude
        max: maximum magnitude
    Output:
        mag: magnitude
    """

    dist = np.linalg.norm(event_loc - station_loc, axis=-1, keepdims=True)
    # mag_ = ( data - 2.48 + 2.76 * np.log10(dist) )
    ## Picozzi et al. (2018) A rapid response magnitude scale...
    c0, c1, c2, c3 = 1.08, 0.93, -0.015, -1.68
    mag_ = (amp - c0 - c3 * np.log10(np.maximum(dist, 0.1))) / c1 + 3.5
    ## Atkinson, G. M. (2015). Ground-Motion Prediction Equation...
    # c0, c1, c2, c3, c4 = (-4.151, 1.762, -0.09509, -1.669, -0.0006)
    # mag_ = (data - c0 - c3*np.log10(dist))/c1
    # mag = np.sum(mag_ * weight) / (np.sum(weight)+1e-6)
    # (Watanabe, 1971) https://www.jstage.jst.go.jp/article/zisin1948/24/3/24_3_189/_pdf/-char/ja
    # mag_ = 1.0/0.85 * (data + 1.73 * np.log10(np.maximum(dist, 0.1)) + 2.50)
    mu = np.sum(mag_ * weight) / (np.sum(weight) + 1e-6)
    std = np.sqrt(np.sum((mag_ - mu) ** 2 * weight) / (np.sum(weight) + 1e-12))
    mask = np.abs(mag_ - mu) <= 2 * std
    mag = np.sum(mag_[mask] * weight[mask]) / (np.sum(weight[mask]) + 1e-6)
    mag = np.clip(mag, min, max)
    return mag


def calc_amp(mag, event_loc, station_loc):
    """
    Calculate amplitude from magnitude and distance
    Input:
        mag: magnitude
        event_loc: event location (n, 3)
        station_loc: station location (n, 3)
    Output:
        logA: log amplitude in log10(cm/s)
    """
    dist = np.linalg.norm(event_loc - station_loc, axis=-1, keepdims=True)
    # logA = mag + 2.48 - 2.76 * np.log10(dist)
    ## Picozzi et al. (2018) A rapid response magnitude scale...
    c0, c1, c2, c3 = 1.08, 0.93, -0.015, -1.68
    logA = c0 + c1 * (mag - 3.5) + c3 * np.log10(np.maximum(dist, 0.1))
    ## Atkinson, G. M. (2015). Ground-Motion Prediction Equation...
    # c0, c1, c2, c3, c4 = (-4.151, 1.762, -0.09509, -1.669, -0.0006)
    # logA = c0 + c1*mag + c3*np.log10(dist)
    # (Watanabe, 1971) https://www.jstage.jst.go.jp/article/zisin1948/24/3/24_3_189/_pdf/-char/ja
    # logA = 0.85 * mag - 2.50 - 1.73 * np.log10(np.maximum(dist, 0.1))
    return logA



# %%
import numpy as np
import scipy
from sklearn.base import BaseEstimator

from ._ransac import RANSACRegressor
from .eikonal2d import grad_traveltime, init_eikonal2d, traveltime
from .gmpe import calc_amp, calc_mag

# from sklearn.linear_model import RANSACRegressor

np.random.seed(0)


class ADLoc(BaseEstimator):
    def __init__(self, config, stations, num_event=1, events=None, eikonal=None):
        """
        events: [x, y, z, t]
        """
        xlim = config["xlim_km"]
        ylim = config["ylim_km"]
        zlim = config["zlim_km"]
        if "vel" in config:
            vel = config["vel"]
        else:
            vel = {"P": 6.0, "S": 6.0 / 1.73}
            if eikonal is None:
                print(f"Using default velocity: {vel}")
        self.config = config
        self.stations = stations
        self.vel = vel

        self.num_event = num_event

        self.eikonal = eikonal

        if events is not None:
            assert events.shape == (num_event, 4)
            self.events = events
        else:
            self.events = np.array(
                [[(xlim[0] + xlim[1]) / 2, (ylim[0] + ylim[1]) / 2, (zlim[0] + zlim[1]) / 2, 0]] * num_event
            )
        if ("use_amplitude" in config) and (config["use_amplitude"]):
            self.use_amplitude = True
            self.magnitudes = np.ones(num_event) * -999.0
        else:
            self.use_amplitude = False

    @staticmethod
    def l2_loss_grad(event, X, y, vel={0: 6.0, 1: 6.0 / 1.75}, stations=None, eikonal=None, reduce_loss=False):
        """
        X: data_frame with columns ["timestamp", "x_km", "y_km", "z_km", "type"]
        """
        station_index = X[:, 0].astype(int)
        phase_type = X[:, 1].astype(int)
        phase_weight = X[:, 2]

        if eikonal is None:
            v = np.array([vel[t] for t in phase_type])
            tt = np.linalg.norm(event[:3] - stations[station_index, :3], axis=-1) / v + event[3]
        else:
            tt = traveltime(0, station_index, phase_type, event[np.newaxis, :3], stations, eikonal) + event[3]

        if reduce_loss:
            loss = 0.5 * np.sum((tt - y) ** 2 * phase_weight)
        else:
            loss = (tt - y) * phase_weight  ## For LSQR

        J = np.ones((len(X), 4))
        if eikonal is None:
            J[:, :3] = (
                (event[:3] - stations[station_index, :3])
                / np.linalg.norm(event[:3] - stations[station_index, :3], axis=-1, keepdims=True)
                / v[:, np.newaxis]
            )
        else:
            grad = grad_traveltime(0, station_index, phase_type, event[np.newaxis, :3], stations, eikonal)
            J[:, :3] = grad

        if reduce_loss:
            J = np.sum((tt - y)[:, np.newaxis] * J * phase_weight[:, np.newaxis], axis=0)
        else:
            J = J * phase_weight[:, np.newaxis]  ## For LSQR

        return loss, J

    @staticmethod
    def huber_loss_grad(event, X, y, vel={0: 6.0, 1: 6.0 / 1.75}, stations=None, eikonal=None, sigma=1.0):

        station_index = X[:, 0].astype(int)
        phase_type = X[:, 1].astype(int)
        phase_weight = X[:, 2]

        if eikonal is None:
            v = np.array([vel[t] for t in phase_type])
            tt = np.linalg.norm(event[:3] - stations[station_index, :3], axis=-1) / v + event[3]
        else:
            tt = traveltime(0, station_index, phase_type, event[np.newaxis, :3], stations, eikonal) + event[3]

        t_diff = tt - y
        l1 = np.squeeze((np.abs(t_diff) > sigma))
        l2 = np.squeeze((np.abs(t_diff) <= sigma))

        loss = np.sum((sigma * np.abs(t_diff[l1]) - 0.5 * sigma**2) * phase_weight[l1]) + np.sum(
            0.5 * t_diff[l2] ** 2 * phase_weight[l2]
        )

        J = np.ones((len(X), 4))
        if eikonal is None:
            J[:, :3] = (
                (event[:3] - stations[station_index, :3])
                / np.linalg.norm(event[:3] - stations[station_index, :3], axis=-1, keepdims=True)
                / v[:, np.newaxis]
            )
        else:
            grad = grad_traveltime(0, station_index, phase_type, event[np.newaxis, :3], stations, eikonal)
            J[:, :3] = grad

        J = np.sum(
            sigma * np.sign(t_diff[l1, np.newaxis]) * J[l1] * phase_weight[l1, np.newaxis], axis=0, keepdims=True
        ) + np.sum(t_diff[l2, np.newaxis] * J[l2] * phase_weight[l2, np.newaxis], axis=0, keepdims=True)

        return loss, J

    def fit(self, X, y=None, event_index=0):

        if self.use_amplitude:
            yt = y[:, 0]  # time
            ya = y[:, 1]  # amplitude
        else:
            yt = y[:, 0]

        opt = scipy.optimize.minimize(
            self.huber_loss_grad,
            # self.l2_loss_grad,
            x0=self.events[event_index],
            method="L-BFGS-B",
            jac=True,
            args=(X, yt, self.vel, self.stations, self.eikonal),
            # bounds=[
            #     (self.config["xlim_km"][0], self.config["xlim_km"][1]),
            #     (self.config["ylim_km"][0], self.config["ylim_km"][1]),
            #     (self.config["zlim_km"][0], self.config["zlim_km"][1]),
            #     (None, None),
            # ],
            bounds=self.config["bfgs_bounds"],
        )

        self.events[event_index, :] = opt.x
        self.is_fitted_ = True

        if self.use_amplitude:
            station_index = X[:, 0].astype(int)
            phase_weight = X[:, 2:3]
            self.magnitudes[event_index] = calc_mag(
                ya[:, np.newaxis], self.events[event_index, :3], self.stations[station_index, :3], phase_weight
            )

        return self

    # def fit(self, X, y=None, event_index=0, damp=1.0, atol=1e-3, max_iter=100): ## For LSQR

    #     if self.use_amplitude:
    #         yt = y[:, 0]  # time
    #         ya = y[:, 1]  # amplitude
    #     else:
    #         yt = y[:, 0]

    #     event0 = self.events[event_index]
    #     delta = np.inf
    #     for i in range(max_iter):
    #         loss, grad = self.l2_loss_grad(event0, X, yt, self.vel, self.stations, self.eikonal)
    #         A, b = grad, loss
    #         dx = scipy.sparse.linalg.lsqr(A, b, damp=damp)[0]
    #         self.events[event_index] -= dx
    #         delta = np.linalg.norm(dx)
    #         if delta < atol:
    #             self.is_fitted_ = True
    #             break

    #     if self.use_amplitude:
    #         station_index = X[:, 0].astype(int)
    #         phase_weight = X[:, 2:3]
    #         self.magnitudes[event_index] = calc_mag(
    #             ya[:, np.newaxis], self.events[event_index, :3], self.stations[station_index, :3], phase_weight
    #         )

    #     return self

    def predict(self, X, event_index=0):
        """
        X: data_frame with columns ["timestamp", "x_km", "y_km", "z_km", "type"]
        """
        station_index = X[:, 0].astype(int)
        phase_type = X[:, 1].astype(int)

        if self.eikonal is None:
            v = np.array([self.vel[t] for t in phase_type])
            tt = (
                np.linalg.norm(self.events[event_index, :3] - self.stations[station_index, :3], axis=-1) / v
                + self.events[event_index, 3]
            )
        else:
            tt = (
                traveltime(event_index, station_index, phase_type, self.events, self.stations, self.eikonal)
                + self.events[event_index, 3]
            )

        if self.use_amplitude:
            amp = calc_amp(
                self.magnitudes[event_index], self.events[event_index, :3], self.stations[station_index, :3]
            ).squeeze()
            return np.array([tt, amp]).T
        else:
            return np.array([tt]).T

    def weight(self, X, event_index=0):
        """
        X: data_frame with columns ["timestamp", "x_km", "y_km", "z_km", "type"]
        """
        station_index = X[:, 0].astype(int)
        phase_type = X[:, 1].astype(int)

        if self.eikonal is None:
            v = np.array([self.vel[t] for t in phase_type])
            tt = np.linalg.norm(self.events[event_index, :3] - self.stations[station_index, :3], axis=-1) / v
        else:
            tt = traveltime(event_index, station_index, phase_type, self.events, self.stations, self.eikonal)

        return np.array([tt]).T

    def score(self, X, y=None, event_index=0):
        """
        X: idx_sta, type, score, amp
        """
        if len(y) <= 1:
            return -np.inf

        output = self.predict(X, event_index)
        if self.use_amplitude:
            tt = output[:, 0]
            amp = output[:, 1]
            yt = y[:, 0]
            ya = y[:, 1]
            R2_t = 1 - np.sum((yt - tt) ** 2) / (np.sum((y[:, 0] - np.mean(y[:, 0])) ** 2) + 1e-6)
            R2_a = 1 - np.sum((ya - amp) ** 2) / (np.sum((y[:, 1] - np.mean(y[:, 1])) ** 2) + 1e-6)
            # weight = X[:, 2]
            # mu_yt = np.average(yt, weights=weight)
            # mu_ya = np.average(ya, weights=weight)
            # R2_t = 1 - np.sum(weight * (yt - tt) ** 2) / (np.sum(weight * (yt - mu_yt) ** 2) + 1e-6)
            # R2_a = 1 - np.sum(weight * (ya - amp) ** 2) / (np.sum(weight * (ya - mu_ya) ** 2) + 1e-6)
            R2 = (R2_t + R2_a) / 2
        else:
            tt = output
            yt = y
            R2 = 1 - np.sum((yt - tt) ** 2) / (np.sum((y - np.mean(y)) ** 2) + 1e-6)
            # weight = X[:, 2]
            # mu_y = np.average(yt, weights=weight)
            # R2 = 1 - np.sum(weight * (yt - tt) ** 2) / (np.sum(weight * (yt - mu_y) ** 2) + 1e-6)

        return R2


# %%
if __name__ == "__main__":

    # %%
    import json
    import os

    import matplotlib.pyplot as plt
    import pandas as pd
    from eikonal2d import _interp, eikonal_solve

    ######################################## Create Synthetic Data #########################################
    np.random.seed(0)
    data_path = "data"
    if not os.path.exists(data_path):
        os.makedirs(data_path, exist_ok=True)
    nx = 10
    ny = 10
    nr = int(np.sqrt(nx**2 + ny**2))
    nz = 10
    h = 20.0
    xgrid = np.arange(0, nx) * h
    ygrid = np.arange(0, ny) * h
    rgrid = np.arange(0, nr) * h
    zgrid = np.arange(0, nz) * h
    xlim = [xgrid[0].item(), xgrid[-1].item()]
    ylim = [ygrid[0].item(), ygrid[-1].item()]
    zlim = [zgrid[0].item(), zgrid[-1].item()]
    eikonal_config = {"nr": nr, "nz": nz, "h": h, "xlim_km": xlim, "ylim_km": ylim, "zlim_km": zlim}
    with open(f"{data_path}/config.json", "w") as f:
        json.dump(eikonal_config, f)
    eikonal_config.update({"rgrid": rgrid, "zgrid": zgrid})
    num_station = 10
    num_event = 50
    stations = []
    for i in range(num_station):
        x = np.random.uniform(xgrid[0], xgrid[-1])
        y = np.random.uniform(ygrid[0], ygrid[-1])
        # z = np.random.uniform(zgrid[0], zgrid[0] + 3 * h)
        z = 0.0
        stations.append({"station_id": f"STA{i:02d}", "x_km": x, "y_km": y, "z_km": z, "dt_s": 0.0})
    stations = pd.DataFrame(stations)
    stations["station_index"] = stations.index
    stations.to_csv(f"{data_path}/stations.csv", index=False)
    events = []
    reference_time = pd.to_datetime("2021-01-01T00:00:00.000")
    for i in range(num_event):
        x = np.random.uniform(xgrid[0], xgrid[-1])
        y = np.random.uniform(ygrid[0], ygrid[-1])
        # z = np.random.uniform(zgrid[0], zgrid[-1])
        z = 0.0
        t = i * 30
        # mag = np.random.uniform(1.0, 5.0)
        mag = 4.0
        events.append(
            {
                "event_id": i,
                "event_time": reference_time + pd.Timedelta(seconds=t),
                "x_km": x,
                "y_km": y,
                "z_km": z,
                "magnitude": mag,
            }
        )
    events = pd.DataFrame(events)
    events["event_index"] = events.index
    events["event_time"] = events["event_time"].apply(lambda x: x.isoformat(timespec="milliseconds"))
    events.to_csv(f"{data_path}/events.csv", index=False)
    vpvs_ratio = 1.73
    vp = np.ones((nr, nz)) * 6.0
    vs = vp / vpvs_ratio

    ## eikonal solver
    up = 1000.0 * np.ones((nr, nz), dtype=np.float64)
    us = 1000.0 * np.ones((nr, nz), dtype=np.float64)
    ir0 = np.around((0.0 - rgrid[0]) / h).astype(np.int64)
    iz0 = np.around((0.0 - zgrid[0]) / h).astype(np.int64)
    up[ir0, iz0] = 0.0
    us[ir0, iz0] = 0.0
    up = eikonal_solve(up, vp, h)
    us = eikonal_solve(us, vs, h)
    up = up.ravel()
    us = us.ravel()

    # %%
    picks = []
    for j, station in stations.iterrows():
        for i, event in events.iterrows():
            r = np.array([np.linalg.norm([event["x_km"] - station["x_km"], event["y_km"] - station["y_km"]])])
            z = np.array([event["z_km"] - station["z_km"]])
            if np.random.rand() < 0.5:
                tt = _interp(up, r, z, rgrid[0], zgrid[0], nr, nz, h)[0]
                picks.append(
                    {
                        "event_id": event["event_id"],
                        "station_id": station["station_id"],
                        "phase_type": "P",
                        "phase_time": pd.to_datetime(event["event_time"]) + pd.Timedelta(seconds=tt),
                        "phase_amplitude": calc_amp(
                            event["magnitude"],
                            event[["x_km", "y_km", "z_km"]].values.astype(np.float32),
                            station[["x_km", "y_km", "z_km"]].values.astype(np.float32),
                        ).item(),
                        "phase_score": 1.0,
                        "travel_time": tt,
                    }
                )
            if np.random.rand() < 0.5:
                tt = _interp(us, r, z, rgrid[0], zgrid[0], nr, nz, h)[0]
                picks.append(
                    {
                        "event_id": event["event_id"],
                        "station_id": station["station_id"],
                        "phase_type": "S",
                        "phase_time": pd.to_datetime(event["event_time"]) + pd.Timedelta(seconds=tt),
                        "phase_amplitude": calc_amp(
                            event["magnitude"],
                            event[["x_km", "y_km", "z_km"]].values.astype(np.float32),
                            station[["x_km", "y_km", "z_km"]].values.astype(np.float32),
                        ).item(),
                        "phase_score": 1.0,
                        "travel_time": tt,
                    }
                )
    picks = pd.DataFrame(picks)
    # use picks,  stations.index, events.index to set station_index and
    picks["phase_time"] = picks["phase_time"].apply(lambda x: x.isoformat(timespec="milliseconds"))
    picks["event_index"] = picks["event_id"].map(events.set_index("event_id")["event_index"])
    picks["station_index"] = picks["station_id"].map(stations.set_index("station_id")["station_index"])
    picks.to_csv(f"{data_path}/picks.csv", index=False)
    # %%
    fig, ax = plt.subplots(1, 2, figsize=(12, 5))
    im = ax[0].imshow(vp[:, :], cmap="viridis")
    fig.colorbar(im, ax=ax[0])
    ax[0].set_title("Vp")
    im = ax[1].imshow(vs[:, :], cmap="viridis")
    fig.colorbar(im, ax=ax[1])
    ax[1].set_title("Vs")
    plt.savefig(f"{data_path}/true2d_vp_vs.png")

    # %%
    fig, ax = plt.subplots(1, 1, squeeze=False, figsize=(5, 5))
    ax[0, 0].scatter(stations["x_km"], stations["y_km"], c=stations["z_km"], marker="^", label="Station")
    ax[0, 0].scatter(events["x_km"], events["y_km"], c=events["z_km"], marker=".", label="Event")
    ax[0, 0].set_xlabel("x (km)")
    ax[0, 0].set_ylabel("y (km)")
    ax[0, 0].legend()
    ax[0, 0].set_title("Station and Event Locations")
    plt.savefig(f"{data_path}/station_event_3d.png")
    # %%
    fig, ax = plt.subplots(3, 1, squeeze=False, figsize=(10, 15))
    picks = picks.merge(stations, on="station_id")
    mapping_color = lambda x: f"C{int(x)}"
    ax[0, 0].scatter(
        pd.to_datetime(picks["phase_time"]),
        picks["x_km"],
        c=picks["event_index"].apply(mapping_color),
        s=20,
    )
    ax[0, 0].scatter(
        pd.to_datetime(events["event_time"]), events["x_km"], c=events["event_index"].apply(mapping_color), marker="x"
    )
    ax[0, 0].set_xlabel("Time (s)")
    ax[0, 0].set_ylabel("x (km)")
    ax[1, 0].scatter(
        pd.to_datetime(picks["phase_time"]),
        picks["y_km"],
        c=picks["event_index"].apply(mapping_color),
        s=20,
    )
    ax[1, 0].scatter(
        pd.to_datetime(events["event_time"]), events["y_km"], c=events["event_index"].apply(mapping_color), marker="x"
    )
    ax[1, 0].set_xlabel("Time (s)")
    ax[1, 0].set_ylabel("y (km)")
    ax[2, 0].scatter(
        pd.to_datetime(picks["phase_time"]),
        picks["z_km"],
        c=picks["event_index"].apply(mapping_color),
        s=20,
    )
    ax[2, 0].scatter(
        pd.to_datetime(events["event_time"]), events["z_km"], c=events["event_index"].apply(mapping_color), marker="x"
    )
    ax[2, 0].set_xlabel("Time (s)")
    ax[2, 0].set_ylabel("z (km)")
    ax[2, 0].invert_yaxis()
    plt.savefig(f"{data_path}/picks_3d.png")

    fig, ax = plt.subplots(1, 1)
    picks = picks.merge(events, on="event_index", suffixes=("_station", "_event"))
    picks["dist_km"] = np.linalg.norm(
        picks[["x_km_station", "y_km_station", "z_km_station"]].values
        - picks[["x_km_event", "y_km_event", "z_km_event"]].values,
        axis=1,
    )
    ax.scatter(np.log10(picks["dist_km"]), picks["phase_amplitude"], c=picks["event_index"].apply(mapping_color), s=30)
    ax.set_xlabel("Distance (km)")
    ax.set_ylabel("Amplitude")
    plt.savefig(f"{data_path}/picks_dist_amp.png")

    # %%
    ######################################### Load Synthetic Data #########################################
    data_path = "data"
    events = pd.read_csv(f"{data_path}/events.csv")
    stations = pd.read_csv(f"{data_path}/stations.csv")
    picks = pd.read_csv(f"{data_path}/picks.csv")
    picks = picks.merge(events[["event_index", "event_time"]], on="event_index")

    #### make the time values relative to event time in seconds
    picks["phase_time_origin"] = picks["phase_time"].copy()
    picks["phase_time"] = (
        pd.to_datetime(picks["phase_time"]) - pd.to_datetime(picks["event_time"])
    ).dt.total_seconds()  # relative to event time (arrival time)
    picks.drop(columns=["event_time"], inplace=True)
    events["event_time_origin"] = events["event_time"].copy()
    events["event_time"] = np.zeros(len(events))  # relative to event time
    ####

    with open(f"{data_path}/config.json", "r") as f:
        eikonal_config = json.load(f)
    num_event = len(events)
    num_station = len(stations)
    ## eikonal solver
    vpvs_ratio = 1.73
    vel = {
        "Z": [0, 100.0],
        "P": [6.0, 6.0],
        "S": [6.0 / vpvs_ratio, 6.0 / vpvs_ratio],
    }
    eikonal_config["vel"] = vel
    eikonal_config = init_eikonal2d(eikonal_config)

    # %%
    stations["idx_sta"] = stations.index  # reindex in case the index does not start from 0 or is not continuous
    events["idx_eve"] = events.index  # reindex in case the index does not start from 0 or is not continuous
    picks = picks.merge(events[["event_index", "idx_eve"]], on="event_index")
    picks = picks.merge(stations[["station_id", "idx_sta"]], on="station_id")

    # %%
    event_index = 1
    num_event = len(events)
    # for idx_eve, picks_event in picks.groupby("idx_eve"):
    picks_event = picks[picks["idx_eve"] == event_index]
    event_loc = events.loc[event_index]
    print(f"Event {event_loc['event_index']} at ({event_loc['x_km']}, {event_loc['y_km']})")
    for _, pick in picks_event.iterrows():
        station_loc = stations.loc[pick["idx_sta"]]
        print(f"Station {station_loc['station_id']} at ({station_loc['x_km']}, {station_loc['y_km']})")

    fig, ax = plt.subplots(1, 2, figsize=(12, 5), squeeze=False)
    tmp = picks_event.merge(stations[["x_km", "y_km", "z_km", "station_id"]], on="station_id")
    tmp["dist_km"] = tmp[["x_km", "y_km", "z_km"]].apply(
        lambda x: np.linalg.norm(x - event_loc[["x_km", "y_km", "z_km"]]), axis=1
    )
    im = ax[0, 0].scatter(tmp["x_km"], tmp["y_km"], c=tmp["phase_time"], cmap="viridis_r", marker="^")
    ax[0, 0].scatter(event_loc["x_km"], event_loc["y_km"], c="r", marker="x")
    ax[0, 0].set_xlabel("x (km)")
    ax[0, 0].set_ylabel("y (km)")
    ax[0, 0].set_aspect("equal")
    plt.colorbar(im, ax=ax[0, 0])

    colors = lambda x: "r" if x == "P" else "b"
    ax[0, 1].scatter(tmp["dist_km"], tmp["phase_time"], c=tmp["phase_type"].apply(colors), s=30)
    ax[0, 1].set_ylabel("Time (s)")
    ax[0, 1].set_xlabel("Distance (km)")
    plt.savefig(f"{data_path}/picks_event_{event_index}.png")

    # %%
    config = {}
    config["xlim_km"] = eikonal_config["xlim_km"]
    config["ylim_km"] = eikonal_config["ylim_km"]
    config["zlim_km"] = eikonal_config["zlim_km"]
    config["bfgs_bounds"] = [
        (config["xlim_km"][0], config["xlim_km"][1]),
        (config["ylim_km"][0], config["ylim_km"][1]),
        (config["zlim_km"][0], config["zlim_km"][1]),
        (None, None),
    ]
    config["vel"] = {"P": 6.0, "S": 6.0 / 1.73}
    config["use_amplitude"] = True
    X = picks_event.merge(
        stations[["x_km", "y_km", "z_km", "station_id"]],
        on="station_id",
    )
    # t0 = X["phase_time"].min() ## already convert to travel time in seconds

    X.rename(
        columns={"phase_type": "type", "phase_time": "t_s", "phase_score": "score", "phase_amplitude": "amp"},
        inplace=True,
    )
    # X["t_s"] = (X["t_s"] - t0).dt.total_seconds()
    X = X[["x_km", "y_km", "z_km", "t_s", "type", "score", "idx_sta", "amp"]]
    mapping_int = {"P": 0, "S": 1}
    config["vel"] = {mapping_int[k]: v for k, v in config["vel"].items()}
    X["type"] = X["type"].apply(lambda x: mapping_int[x.upper()])

    estimator = ADLoc(
        config,
        # events=events[["x_km", "y_km", "z_km", "event_time"]].values + np.random.rand(num_event, 4) * 10,
        stations=stations[["x_km", "y_km", "z_km"]].values,
        num_event=num_event,
        eikonal=eikonal_config,
    )
    output = estimator.predict(X[["idx_sta", "type"]].values, event_index=event_index)
    tt_init = output[:, 0]
    amp_init = output[:, 1]
    estimator.score(X[["idx_sta", "type"]].values, y=X[["t_s", "amp"]].values, event_index=event_index)

    print(f"Init event loc: {estimator.events[event_index].round(3)}")

    # %%
    estimator.fit(X[["idx_sta", "type", "score", "amp"]].values, y=X[["t_s", "amp"]].values, event_index=event_index)
    estimator.score(X[["idx_sta", "type"]].values, y=X[["t_s", "amp"]].values, event_index=event_index)
    output = estimator.predict(X[["idx_sta", "type"]].values, event_index=event_index)
    tt = output[:, 0]
    amp = output[:, 1]
    print("Basic:")
    print(f"True event loc: {event_loc[['x_km', 'y_km', 'z_km']].values.astype(float).round(3)}")
    print(f"Invt event loc: {estimator.events[event_index].round(3)}")

    # %%
    fig, ax = plt.subplots(1, 2, figsize=(10, 5), squeeze=False)
    X["dist_km"] = X[["x_km", "y_km", "z_km"]].apply(
        lambda x: np.linalg.norm(x - estimator.events[event_index, :3]), axis=1
    )
    colors = lambda x: "r" if x == 0 else "b"
    ax[0, 0].scatter(X["dist_km"], X["t_s"], s=20, marker="o", label="Picks")
    ax[0, 0].scatter(X["dist_km"], tt_init, s=40, marker="x", label="Init")
    ax[0, 0].scatter(X["dist_km"], tt, s=40, marker="x", label="Invert")
    ax[0, 0].set_xlabel("Distance (km)")
    ax[0, 0].set_ylabel("Time (s)")
    ax[0, 0].legend()

    ax[0, 1].scatter(np.log10(X["dist_km"]), X["amp"], s=20, marker="o", label="Picks")
    # ax[0, 1].scatter(np.log10(X["dist_km"]), amp_init, s=40, marker="x", label="Init")
    ax[0, 1].scatter(np.log10(X["dist_km"]), amp, s=40, marker="x", label="Invert")
    ax[0, 1].set_xlabel("Distance (km)")
    ax[0, 1].set_ylabel("Amplitude")
    ax[0, 1].legend()
    fig.savefig(f"{data_path}/picks_xt_{event_index}_adloc.png")

    # %%
    reg = RANSACRegressor(
        estimator=ADLoc(
            config, stations=stations[["x_km", "y_km", "z_km"]].values, num_event=num_event, eikonal=eikonal_config
        ),
        random_state=0,
        min_samples=4,
        residual_threshold=[1.0, 0.5],
    )
    reg.fit(X[["idx_sta", "type", "score", "amp"]].values, X[["t_s", "amp"]].values, event_index=event_index)
    mask_invt = reg.inlier_mask_
    estimator = reg.estimator_
    estimator.score(X[["idx_sta", "type"]].values, y=X[["t_s", "amp"]].values, event_index=event_index)
    output = estimator.predict(X[["idx_sta", "type"]].values, event_index=event_index)
    tt = output[:, 0]
    amp = output[:, 1]
    print("RANSAC:")
    print(f"True event loc: {event_loc[['x_km', 'y_km', 'z_km']].values.astype(float).round(3)}")
    print(f"Invt event loc: {estimator.events[event_index].round(3)}")

    # %% add random picks
    config["vel"] = {"P": 6.0, "S": 6.0 / 1.73}
    X = picks_event.merge(stations[["x_km", "y_km", "z_km", "station_id"]], on="station_id")
    # t0 = X["phase_time"].min() ## already convert to travel time in seconds
    X.rename(
        columns={"phase_type": "type", "phase_time": "t_s", "phase_score": "score", "phase_amplitude": "amp"},
        inplace=True,
    )
    # X["t_s"] = (X["t_s"] - t0).dt.total_seconds()
    X = X[["x_km", "y_km", "z_km", "t_s", "type", "score", "idx_sta", "amp"]]
    mapping_int = {"P": 0, "S": 1}
    config["vel"] = {mapping_int[k]: v for k, v in config["vel"].items()}
    X["type"] = X["type"].apply(lambda x: mapping_int[x.upper()])

    num_noise = len(X) * 2
    noise = pd.DataFrame(
        {
            "x_km": np.random.rand(num_noise) * (X["x_km"].max() - X["x_km"].min()) + X["x_km"].min(),
            "y_km": np.random.rand(num_noise) * (X["y_km"].max() - X["y_km"].min()) + X["y_km"].min(),
            "z_km": np.random.rand(num_noise) * (X["z_km"].max() - X["z_km"].min()) + X["z_km"].min(),
            "t_s": np.random.rand(num_noise) * (X["t_s"].max() - X["t_s"].min()) + X["t_s"].min(),
            "type": np.random.choice([0, 1], num_noise),
            "score": 1.0,
            "idx_sta": np.random.choice(X["idx_sta"], num_noise),
            "amp": np.random.rand(num_noise) * (X["amp"].max() - X["amp"].min()) * 2 + X["amp"].min(),
            "mask": [0] * num_noise,
        }
    )
    X["mask"] = 1
    X = pd.concat([X, noise], ignore_index=True)
    X["dist_km"] = X[["x_km", "y_km", "z_km"]].apply(
        lambda x: np.linalg.norm(x - event_loc[["x_km", "y_km", "z_km"]]), axis=1
    )

    # %%
    fig, ax = plt.subplots(1, 2, figsize=(10, 5), squeeze=False)
    colors = lambda x: "r" if x == 0 else "b"
    # ax.scatter(X["t_s"], X["dist_km"], c=X["type"].apply(colors), marker="o", label="Picks")
    # ax.scatter(X["t_s"][X["mask"] == 0], X["dist_km"][X["mask"] == 0], c="k", marker="o", alpha=0.6, label="Noise")
    ax[0, 0].scatter(X["dist_km"], X["t_s"], c=X["type"].apply(colors), s=30, marker="o", label="Picks")
    ax[0, 0].scatter(
        X["dist_km"][X["mask"] == 0], X["t_s"][X["mask"] == 0], c="k", s=40, marker="o", alpha=0.6, label="Noise"
    )
    ax[0, 0].set_ylabel("Time (s)")
    ax[0, 0].set_xlabel("Distance (km)")
    ax[0, 0].legend()

    ax[0, 1].scatter(np.log10(X["dist_km"]), X["amp"], c=X["type"].apply(colors), s=30, marker="o", label="Picks")
    ax[0, 1].scatter(
        np.log10(X["dist_km"][X["mask"] == 0]), X["amp"][X["mask"] == 0], c="k", s=30, marker="o", label="Noise"
    )
    ax[0, 1].set_xlabel("Distance (km)")
    ax[0, 1].set_ylabel("Amplitude")
    plt.savefig(f"{data_path}/picks_xt_{event_index}_noise.png")

    # %%
    reg = RANSACRegressor(
        estimator=ADLoc(
            config, stations=stations[["x_km", "y_km", "z_km"]].values, num_event=num_event, eikonal=eikonal_config
        ),
        random_state=0,
        min_samples=4,
        residual_threshold=[1.0, 0.5],
    )
    reg.fit(X[["idx_sta", "type", "score", "amp"]].values, X[["t_s", "amp"]].values, event_index=event_index)
    mask_invt = reg.inlier_mask_
    estimator = reg.estimator_
    estimator.score(X[["idx_sta", "type"]].values, y=X[["t_s", "amp"]].values, event_index=event_index)
    output = estimator.predict(X[["idx_sta", "type"]].values, event_index=event_index)
    tt = output[:, 0]
    amp = output[:, 1]
    print("RANSAC with noise picks:")
    print(f"True event loc: {event_loc[['x_km', 'y_km', 'z_km']].values.astype(float).round(3)}")
    print(f"Invt event loc: {estimator.events[event_index].round(3)}")

    # %%
    fig, ax = plt.subplots(2, 2, figsize=(8, 8), squeeze=False)
    X["dist_km"] = X[["x_km", "y_km", "z_km"]].apply(
        lambda x: np.linalg.norm(x - event_loc[["x_km", "y_km", "z_km"]]),
        axis=1,
    )
    colors = lambda x: "r" if x == 0 else "b"
    ax[0, 0].scatter(X["dist_km"], X["t_s"], c=X["type"].apply(colors), s=30, marker="o", label="Picks")
    mask_true = X["mask"] == 1
    ax[0, 0].scatter(X["dist_km"][~mask_true], X["t_s"][~mask_true], s=30, c="k", marker="o", alpha=0.6, label="Noise")
    ax[0, 0].scatter(X["dist_km"][mask_invt], tt[mask_invt], s=50, c="g", marker="x", label="Invert")
    ax[0, 0].set_ylabel("Time (s)")
    ax[0, 0].set_xlabel("Distance (km)")
    ax[0, 0].legend()

    ax[0, 1].scatter(np.log10(X["dist_km"]), X["amp"], c=X["type"].apply(colors), s=30, marker="o", label="Picks")
    ax[0, 1].scatter(np.log10(X["dist_km"][~mask_true]), X["amp"][~mask_true], c="k", s=30, marker="o", label="Noise")
    ax[0, 1].scatter(np.log10(X["dist_km"][mask_invt]), amp[mask_invt], c="g", s=50, marker="x", label="Invert")

    X["dist_km"] = X[["x_km", "y_km", "z_km"]].apply(
        lambda x: np.linalg.norm(x - estimator.events[event_index, :3]), axis=1
    )
    colors = lambda x: "r" if x == 0 else "b"
    ax[1, 0].scatter(X["dist_km"], X["t_s"], c=X["type"].apply(colors), s=30, marker="o", label="Picks")
    mask_true = X["mask"] == 1
    ax[1, 0].scatter(X["dist_km"][~mask_true], X["t_s"][~mask_true], s=30, c="k", marker="o", alpha=0.6, label="Noise")
    ax[1, 0].scatter(X["dist_km"][mask_invt], tt[mask_invt], s=50, c="g", marker="x", label="Invert")
    ax[1, 0].set_ylabel("Time (s)")
    ax[1, 0].set_xlabel("Distance (km)")
    ax[1, 0].legend()

    ax[1, 1].scatter(np.log10(X["dist_km"]), X["amp"], c=X["type"].apply(colors), s=30, marker="o", label="Picks")
    ax[1, 1].scatter(np.log10(X["dist_km"][~mask_true]), X["amp"][~mask_true], c="k", s=30, marker="o", label="Noise")
    ax[1, 1].scatter(np.log10(X["dist_km"][mask_invt]), amp[mask_invt], c="g", s=50, marker="x", label="Invert")

    fig.savefig(f"{data_path}/picks_xt_{event_index}_ransac.png")



import numpy as np
import torch
from sklearn.neighbors import NearestNeighbors
from torch.utils.data import Dataset
from tqdm import tqdm

# class PhaseDataset(Dataset):
#     def __init__(self, picks, events, stations, batch_size=100, double_difference=False, config=None):
#         self.picks = picks
#         self.events = events
#         self.stations = stations
#         self.config = config
#         self.double_difference = double_difference
#         self.batch_size = batch_size
#         if double_difference:
#             self.read_data_dd()
#         else:
#             self.read_data()

#     def __len__(self):
#         ## TODO: return batch
#         return 1
#         # return (len(self.events) - 1) // self.batch_size + 1

#     def read_data(self):
#         event_index = []
#         station_index = []
#         phase_score = []
#         phase_time = []
#         phase_type = []

#         picks_by_event = self.picks.groupby("index")
#         for key, group in picks_by_event:
#             if key == -1:
#                 continue
#             phase_time.append(group["phase_time"].values)
#             phase_score.append(group["phase_score"].values)
#             phase_type.extend(group["phase_type"].values.tolist())
#             event_index.extend([key] * len(group))
#             station_index.append(self.stations.loc[group["station_id"], "index"].values)

#         phase_time = np.concatenate(phase_time)
#         phase_score = np.concatenate(phase_score)
#         phase_type = np.array([{"P": 0, "S": 1}[x.upper()] for x in phase_type])
#         event_index = np.array(event_index)
#         station_index = np.concatenate(station_index)

#         self.station_index = torch.tensor(station_index, dtype=torch.long)
#         self.event_index = torch.tensor(event_index, dtype=torch.long)
#         self.phase_weight = torch.tensor(phase_score, dtype=torch.float32)
#         self.phase_time = torch.tensor(phase_time, dtype=torch.float32)
#         self.phase_type = torch.tensor(phase_type, dtype=torch.long)

#     def read_data_dd(self):
#         event_index = []
#         station_index = []
#         phase_score = []
#         phase_time = []
#         phase_type = []

#         event_loc = self.events[["x_km", "y_km", "z_km"]].values
#         event_time = self.events["time"].values[:, np.newaxis]

#         neigh = NearestNeighbors(radius=self.config.min_pair_dist, metric="euclidean")
#         neigh.fit(event_loc)

#         picks_by_event = self.picks.groupby("index")

#         for key1, group1 in tqdm(picks_by_event, total=len(picks_by_event), desc="Generating pairs"):
#             if key1 == -1:
#                 continue

#             for key2 in neigh.radius_neighbors([event_loc[key1]], return_distance=False)[0]:
#                 if key1 >= key2:
#                     continue

#                 common = group1.merge(picks_by_event.get_group(key2), on=["station_id", "phase_type"], how="inner")
#                 phase_time.append(common["phase_time_x"].values - common["phase_time_y"].values)
#                 phase_score.append(common["phase_score_x"].values * common["phase_score_y"].values)
#                 phase_type.extend(common["phase_type"].values.tolist())
#                 event_index.extend([[key1, key2]] * len(common))
#                 station_index.append(self.stations.loc[common["station_id"], "index"].values)

#         if len(phase_time) > 0:
#             phase_time = np.concatenate(phase_time)
#             phase_score = np.concatenate(phase_score)
#             phase_type = np.array([{"P": 0, "S": 1}[x.upper()] for x in phase_type])
#             event_index = np.array(event_index)
#             station_index = np.concatenate(station_index)

#             # %%
#             self.station_index = torch.tensor(station_index, dtype=torch.long)
#             self.event_index = torch.tensor(event_index, dtype=torch.long)
#             self.phase_weight = torch.tensor(phase_score, dtype=torch.float32)
#             self.phase_time = torch.tensor(phase_time, dtype=torch.float32)
#             self.phase_type = torch.tensor(phase_type, dtype=torch.long)
#         else:
#             self.station_index = torch.tensor([], dtype=torch.long)
#             self.event_index = torch.tensor([], dtype=torch.long)
#             self.phase_weight = torch.tensor([], dtype=torch.float32)
#             self.phase_time = torch.tensor([], dtype=torch.float32)
#             self.phase_type = torch.tensor([], dtype=torch.long)

#     def __getitem__(self, i):
#         # phase_time = self.picks[self.picks["event_index"] == self.events.loc[i, "event_index"]]["phase_time"].values
#         # phase_score = self.picks[self.picks["event_index"] == self.events.loc[i, "event_index"]]["phase_score"].values
#         # phase_type = self.picks[self.picks["event_index"] == self.events.loc[i, "event_index"]][
#         #     "phase_type"
#         # ].values.tolist()
#         # event_index = np.array([i] * len(self.picks[self.picks["event_index"] == self.events.loc[i, "event_index"]]))
#         # station_index = self.stations.loc[
#         #     self.picks[self.picks["event_index"] == self.events.loc[i, "event_index"]]["station_id"], "index"
#         # ].values

#         return {
#             "event_index": self.event_index,
#             "station_index": self.station_index,
#             "phase_time": self.phase_time,
#             "phase_weight": self.phase_weight,
#             "phase_type": self.phase_type,
#         }


class PhaseDataset(Dataset):
    def __init__(
        self,
        picks,
        events,
        stations,
        batch_size=1000,
        config=None,
        rank=0,
        world_size=1,
    ):
        self.picks = picks
        self.events = events
        self.stations = stations
        self.config = config
        self.batch_size = batch_size

        # preprocess
        self.picks_by_event = picks.groupby("idx_eve")
        self.idx_eve_batch = np.array_split(self.events["idx_eve"], (len(self.events) - 1) // self.batch_size + 1)[
            rank::world_size
        ]
        self.read_data()

    def __len__(self):
        return len(self.idx_eve_batch)

    def read_data(self):
        meta = {}
        for i, index_batch in enumerate(self.idx_eve_batch):
            idx_eve = []
            idx_sta = []
            phase_score = []
            phase_time = []
            phase_type = []

            for key in index_batch:
                if key == -1:
                    continue
                group = self.picks_by_event.get_group(key)
                phase_time.append(group["travel_time"].values)  # seconds
                # phase_time.append(group["phase_time"].values) # datetime not supported
                phase_score.append(group["phase_score"].values)
                phase_type.extend(group["phase_type"].values.tolist())
                idx_eve.extend(group["idx_eve"].values.tolist())
                idx_sta.append(group["idx_sta"].values.tolist())

            phase_time = np.concatenate(phase_time)
            phase_score = np.concatenate(phase_score)
            # phase_type = np.array([{"P": 0, "S": 1}[x.upper()] for x in phase_type])
            # if ("P" in phase_type) and ("S" in phase_type):
            #     phase_type = np.array([{"P": 0, "S": 1}[x] for x in phase_type])
            idx_eve = np.array(idx_eve)
            idx_sta = np.concatenate(idx_sta)

            idx_sta = torch.tensor(idx_sta, dtype=torch.long)
            idx_eve = torch.tensor(idx_eve, dtype=torch.long)
            phase_weight = torch.tensor(phase_score, dtype=torch.float32)
            phase_time = torch.tensor(phase_time, dtype=torch.float32)
            phase_type = np.array(phase_type)

            meta[i] = {
                "idx_eve": idx_eve,
                "idx_sta": idx_sta,
                "phase_time": phase_time,
                "phase_weight": phase_weight,
                "phase_type": phase_type,
            }
        self.meta = meta

    def __getitem__(self, i):
        return self.meta[i]


class PhaseDatasetDT(Dataset):
    def __init__(
        self,
        pairs,
        picks,
        events,
        stations,
        batch_size=10000,
        valid_index=None,
        config=None,
        rank=0,
        world_size=1,
    ):
        self.pairs = pairs
        self.picks = picks
        self.events = events
        self.stations = stations
        self.config = config
        self.batch_size = batch_size
        if valid_index is None:
            valid_index = np.ones(len(self.pairs), dtype=bool)
        self.valid_index = valid_index

        self.idx_batch = np.array_split(np.arange(len(self.pairs)), (len(self.pairs) - 1) // self.batch_size + 1)[
            rank::world_size
        ]

        print(f"Generated {len(self.pairs)} pairs")
        print(f"Split into {len(self.idx_batch)} batches of size {self.batch_size} at rank {rank}/{world_size}")

    def __len__(self):
        return len(self.idx_batch)

    def __getitem__(self, i):

        idx = self.idx_batch[i][self.valid_index[self.idx_batch[i]]]

        if len(idx) == 0:
            return None  # skip empty batch

        idx1_eve = self.pairs["idx_eve1"][idx].values
        idx2_eve = self.pairs["idx_eve2"][idx].values
        idx_eve = np.stack([idx1_eve, idx2_eve], axis=1)
        idx_sta = self.pairs["idx_sta"][idx].values
        phase_weight = self.pairs["phase_score"][idx].values
        phase_type = self.pairs["phase_type"][idx].values
        phase_time = self.pairs["phase_dtime"][idx].values

        return {
            "idx_eve": torch.tensor(idx_eve, dtype=torch.long),
            "idx_sta": torch.tensor(idx_sta, dtype=torch.long),
            "phase_type": phase_type,
            "phase_weight": torch.tensor(phase_weight, dtype=torch.float32),
            "phase_time": torch.tensor(phase_time, dtype=torch.float32),
        }


class PhaseDatasetDTCC(PhaseDatasetDT):

    def __getitem__(self, i):

        idx = self.idx_batch[i][self.valid_index[self.idx_batch[i]]]
        idx1_eve = self.pairs["idx_eve1"][idx].values
        idx2_eve = self.pairs["idx_eve2"][idx].values
        idx_eve = np.stack([idx1_eve, idx2_eve], axis=1)
        idx_sta = self.pairs["idx_sta"][idx].values
        phase_weight = self.pairs["weight"][idx].values
        phase_type = self.pairs["phase_type"][idx].values
        phase_time = self.pairs["dt"][idx].values

        return {
            "idx_eve": torch.tensor(idx_eve, dtype=torch.long),
            "idx_sta": torch.tensor(idx_sta, dtype=torch.long),
            "phase_type": phase_type,
            "phase_weight": torch.tensor(phase_weight, dtype=torch.float32),
            "phase_time": torch.tensor(phase_time, dtype=torch.float32),
        }



import torch
import torch.distributed as dist
import torch.optim as optim
from tqdm import tqdm


def optimize(args, config, data_loader, data_loader_dd, travel_time):
    if (args.opt.lower() == "lbfgs") or (args.opt.lower() == "bfgs"):
        optimizer = optim.LBFGS(params=travel_time.parameters(), max_iter=1000, line_search_fn="strong_wolfe")
    elif args.opt.lower() == "adam":
        optimizer = optim.Adam(params=travel_time.parameters(), lr=0.1)
    elif args.opt.lower() == "sgd":
        optimizer = optim.SGD(params=travel_time.parameters(), lr=10.0)
    else:
        raise ValueError(f"Unknown optimizer: {args.opt}")

    # init loss
    loss = 0
    loss_dd = 0
    for meta in data_loader:
        station_index = meta["station_index"]
        event_index = meta["event_index"]
        phase_time = meta["phase_time"]
        phase_type = meta["phase_type"]
        phase_weight = meta["phase_weight"]
        loss += travel_time(
            station_index,
            event_index,
            phase_type,
            phase_time,
            phase_weight,
            double_difference=False,
        )["loss"]
        if args.distributed:
            dist.barrier()
            dist.all_reduce(loss)

        if args.double_difference:
            for meta in data_loader_dd:
                station_index = meta["station_index"]
                event_index = meta["event_index"]
                phase_time = meta["phase_time"]
                phase_type = meta["phase_type"]
                phase_weight = meta["phase_weight"]

                loss_dd += travel_time(
                    station_index,
                    event_index,
                    phase_type,
                    phase_time,
                    phase_weight,
                    double_difference=True,
                )["loss"]
                if args.distributed:
                    dist.barrier()
                    dist.all_reduce(loss_dd)

    print(f"Init loss: {loss+loss_dd}:  {loss} + {loss_dd}")

    if (args.opt.lower() == "lbfgs") or (args.opt.lower() == "bfgs"):
        prev_loss = 0
        for i in range(args.epochs):

            def closure():
                optimizer.zero_grad()
                for meta in data_loader:
                    station_index = meta["station_index"]
                    event_index = meta["event_index"]
                    phase_time = meta["phase_time"]
                    phase_type = meta["phase_type"]
                    phase_weight = meta["phase_weight"]
                    loss = travel_time(
                        station_index,
                        event_index,
                        phase_type,
                        phase_time,
                        phase_weight,
                        double_difference=False,
                    )["loss"]
                    if args.distributed:
                        dist.barrier()
                        dist.all_reduce(loss)
                    loss.backward()

                    if args.double_difference:
                        for meta in data_loader_dd:
                            station_index = meta["station_index"]
                            event_index = meta["event_index"]
                            phase_time = meta["phase_time"]
                            phase_type = meta["phase_type"]
                            phase_weight = meta["phase_weight"]

                            loss_dd = travel_time(
                                station_index,
                                event_index,
                                phase_type,
                                phase_time,
                                phase_weight,
                                double_difference=True,
                            )["loss"]
                            if args.distributed:
                                dist.barrier()
                                dist.all_reduce(loss_dd)

                            (loss_dd * args.dd_weight).backward()

                return loss

            optimizer.step(closure)

            loss = 0
            loss_dd = 0
            for meta in data_loader:
                station_index = meta["station_index"]
                event_index = meta["event_index"]
                phase_time = meta["phase_time"]
                phase_type = meta["phase_type"]
                phase_weight = meta["phase_weight"]
                loss += travel_time(
                    station_index,
                    event_index,
                    phase_type,
                    phase_time,
                    phase_weight,
                    double_difference=False,
                )["loss"]
            if args.double_difference:
                for meta in data_loader_dd:
                    station_index = meta["station_index"]
                    event_index = meta["event_index"]
                    phase_time = meta["phase_time"]
                    phase_type = meta["phase_type"]
                    phase_weight = meta["phase_weight"]

                    loss_dd = travel_time(
                        station_index,
                        event_index,
                        phase_type,
                        phase_time,
                        phase_weight,
                        double_difference=True,
                    )["loss"]
            if abs((loss + loss_dd) - prev_loss) < 1e-6:
                break
            prev_loss = loss + loss_dd
            print(f"Invert loss: {loss+loss_dd}:  {loss} + {loss_dd}")

            # set variable range
            travel_time.event_loc.weight.data[:, 2] += (
                (torch.rand_like(travel_time.event_loc.weight.data[:, 2]) - 0.5) * (args.epochs - 1 - i) / args.epochs
            )
            travel_time.event_loc.weight.data[:, 2].clamp_(min=config["zlim_km"][0], max=config["zlim_km"][1])
            # travel_time.event_loc.weight.data[:, 2].clamp_(min=config["mindepth"], max=config["maxdepth"])

    else:
        for i in range(args.epochs):
            optimizer.zero_grad()

            prev_loss = 0
            loss = 0
            loss_dd = 0
            for meta in data_loader:
                station_index = meta["station_index"]
                event_index = meta["event_index"]
                phase_time = meta["phase_time"]
                phase_type = meta["phase_type"]
                phase_weight = meta["phase_weight"]

                loss = travel_time(
                    station_index,
                    event_index,
                    phase_type,
                    phase_time,
                    phase_weight,
                    double_difference=False,
                )["loss"]
                loss.backward()

            if args.double_difference:
                for meta in data_loader_dd:
                    station_index = meta["station_index"]
                    event_index = meta["event_index"]
                    phase_time = meta["phase_time"]
                    phase_type = meta["phase_type"]
                    phase_weight = meta["phase_weight"]

                    loss_dd = travel_time(
                        station_index,
                        event_index,
                        phase_type,
                        phase_time,
                        phase_weight,
                        double_difference=True,
                    )["loss"]
                    (loss_dd * args.dd_weight).backward()

            optimizer.step()

            if abs((loss + loss_dd) - prev_loss) < 1e-3:
                print(f"Loss: {loss+loss_dd}:  {loss} + {loss_dd}")
                break
            prev_loss = loss + loss_dd

            if i % 100 == 0:
                print(f"Loss: {loss+loss_dd}:  {loss} + {loss_dd}")

            # set variable range
            travel_time.event_loc.weight.data[:, 2] += (
                torch.randn_like(travel_time.event_loc.weight.data[:, 2]) * (args.epochs - i) / args.epochs
            )
            travel_time.event_loc.weight.data[:, 2].clamp_(min=config["z(km)"][0], max=config["z(km)"][1])


def optimize_dd(data_loader, travel_time, config):

    # init loss
    loss = 0
    for meta in data_loader:
        station_index = meta["idx_sta"]
        event_index = meta["idx_eve"]
        phase_time = meta["phase_time"]
        phase_type = meta["phase_type"]
        phase_weight = meta["phase_weight"]

        loss += travel_time(
            station_index,
            event_index,
            phase_type,
            phase_time,
            phase_weight,
        )["loss"]

    # if args.distributed:
    #     dist.barrier()
    #     dist.all_reduce(loss)
    print(f"Init loss: {loss}")

    ## invert loss
    # optimizer = optim.Adam(params=travel_time.parameters(), lr=0.1)
    # EPOCHS = 3
    # for i in range(EPOCHS):
    #     loss = 0
    #     optimizer.zero_grad()
    #     for meta in tqdm(data_loader, desc=f"Epoch {i}"):
    #         station_index = meta["idx_sta"]
    #         event_index = meta["idx_eve"]
    #         phase_time = meta["phase_time"]
    #         phase_type = meta["phase_type"]
    #         phase_weight = meta["phase_weight"]

    #         loss_ = travel_time(
    #             station_index,
    #             event_index,
    #             phase_type,
    #             phase_time,
    #             phase_weight,
    #         )["loss"]

    #         # optimizer.zero_grad()
    #         loss_.backward()
    #         # optimizer.step()
    #         loss += loss_

    #     optimizer.step()
    #     print(f"Loss: {loss}")

    optimizer = optim.LBFGS(params=travel_time.parameters(), max_iter=100, line_search_fn="strong_wolfe")

    def closure():
        optimizer.zero_grad()
        loss = 0
        for meta in tqdm(data_loader, desc=f"BFGS"):
            station_index = meta["idx_sta"]
            event_index = meta["idx_eve"]
            phase_time = meta["phase_time"]
            phase_type = meta["phase_type"]
            phase_weight = meta["phase_weight"]

            loss_ = travel_time(
                station_index,
                event_index,
                phase_type,
                phase_time,
                phase_weight,
            )["loss"]
            loss_.backward()
            loss += loss_

        print(f"Loss: {loss}")
        # if args.distributed:
        #     dist.barrier()
        #     dist.all_reduce(loss)

        return loss

    optimizer.step(closure)

    # travel_time.event_loc.weight.grad.data -= travel_time.event_loc.weight.grad.data.mean()
    # travel_time.event_time.weight.grad.data -= travel_time.event_time.weight.grad.data.mean()

    ## updated loss
    loss = 0
    for meta in data_loader:
        station_index = meta["idx_sta"]
        event_index = meta["idx_eve"]
        phase_time = meta["phase_time"]
        phase_type = meta["phase_type"]
        phase_weight = meta["phase_weight"]

        loss += travel_time(
            station_index,
            event_index,
            phase_type,
            phase_time,
            phase_weight,
        )["loss"]

    # if args.distributed:
    #     dist.barrier()
    #     dist.all_reduce(loss)

    print(f"Invert loss: {loss}")

    # set variable range
    # travel_time.event_loc.weight.data[:, 2] += (
    #     (torch.rand_like(travel_time.event_loc.weight.data[:, 2]) - 0.5) * (args.epochs - 1 - i) / args.epochs
    # )
    travel_time.event_loc.weight.data[:, 2].clamp_(min=config["zlim_km"][0], max=config["zlim_km"][1])



import itertools
import shelve
import time
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import scipy.optimize
from numba import njit
from numba.typed import List

np.random.seed(0)

###################################### Eikonal Solver ######################################

# |\nabla u| = f
# ((u - a1)^+)^2 + ((u - a2)^+)^2 + ((u - a3)^+)^2 = f^2 h^2


@njit
def calculate_unique_solution(a, b, f, h):
    d = abs(a - b)
    if d >= f * h:
        return min([a, b]) + f * h
    else:
        return (a + b + np.sqrt(2.0 * f * f * h * h - (a - b) ** 2)) / 2.0


@njit
def sweeping_over_I_J_K(u, I, J, f, h):
    m = len(I)
    n = len(J)

    for i in I:
        for j in J:
            if i == 0:
                uxmin = u[i + 1, j]
            elif i == m - 1:
                uxmin = u[i - 1, j]
            else:
                uxmin = min([u[i - 1, j], u[i + 1, j]])

            if j == 0:
                uymin = u[i, j + 1]
            elif j == n - 1:
                uymin = u[i, j - 1]
            else:
                uymin = min([u[i, j - 1], u[i, j + 1]])

            u_new = calculate_unique_solution(uxmin, uymin, f[i, j], h)

            u[i, j] = min([u_new, u[i, j]])

    return u


@njit
def sweeping(u, v, h):
    f = 1.0 / v  ## slowness

    m, n = u.shape
    I = np.arange(m)
    iI = I[::-1]
    J = np.arange(n)
    iJ = J[::-1]

    u = sweeping_over_I_J_K(u, I, J, f, h)
    u = sweeping_over_I_J_K(u, iI, J, f, h)
    u = sweeping_over_I_J_K(u, iI, iJ, f, h)
    u = sweeping_over_I_J_K(u, I, iJ, f, h)

    return u


def eikonal_solve(u, v, h):
    print("Eikonal Solver: ")
    t0 = time.time()
    for i in range(50):
        u_old = np.copy(u)
        u = sweeping(u, v, h)

        err = np.max(np.abs(u - u_old))
        print(f"Iter {i}, error = {err:.3f}")
        if err < 1e-6:
            break
    print(f"Time: {time.time() - t0:.3f}")
    return u


###################################### Traveltime based on Eikonal Timetable ######################################
@njit
def _get_index(ir, iz, nr, nz, order="C"):
    if order == "C":
        return ir * nz + iz
    elif order == "F":
        return iz * nr + ir
    else:
        raise ValueError("order must be either C or F")


def test_get_index():
    vr, vz = np.meshgrid(np.arange(10), np.arange(20), indexing="ij")
    vr = vr.ravel()
    vz = vz.ravel()
    nr = 10
    nz = 20
    for ir in range(nr):
        for iz in range(nz):
            assert vr[_get_index(ir, iz, nr, nz)] == ir
            assert vz[_get_index(ir, iz, nr, nz)] == iz


@njit
def _interp(time_table, r, z, rgrid0, zgrid0, nr, nz, h):
    shape = r.shape
    r = r.reshape(-1)
    z = z.reshape(-1)

    ir0 = np.floor((r - rgrid0) / h).clip(0, nr - 2).astype(np.int64)
    iz0 = np.floor((z - zgrid0) / h).clip(0, nz - 2).astype(np.int64)
    ir1 = ir0 + 1
    iz1 = iz0 + 1

    ## https://en.wikipedia.org/wiki/Bilinear_interpolation
    r0 = ir0 * h + rgrid0
    r1 = ir1 * h + rgrid0
    z0 = iz0 * h + zgrid0
    z1 = iz1 * h + zgrid0

    Q00 = time_table[_get_index(ir0, iz0, nr, nz)]
    Q01 = time_table[_get_index(ir0, iz1, nr, nz)]
    Q10 = time_table[_get_index(ir1, iz0, nr, nz)]
    Q11 = time_table[_get_index(ir1, iz1, nr, nz)]

    t = (
        1.0
        / (r1 - r0)
        / (z1 - z0)
        * (
            Q00 * (r1 - r) * (z1 - z)
            + Q10 * (r - r0) * (z1 - z)
            + Q01 * (r1 - r) * (z - z0)
            + Q11 * (r - r0) * (z - z0)
        )
    )

    return t.reshape(shape)


# def traveltime(event_loc, station_loc, phase_type, eikonal):
def traveltime(event_index, station_index, phase_type, events, stations, eikonal, vel={0: 6.0, 1: 6.0 / 1.73}):
    """
    event_index: list of event index
    station_index: list of station index
    phase_type: list of phase type
    events: list of event location
    stations: list of station location
    """
    if eikonal is None:
        v = np.array([vel[x] for x in phase_type])
        tt = np.linalg.norm(events[event_index] - stations[station_index], axis=-1, keepdims=False) / v
    else:
        if isinstance(event_index, int):
            event_index = np.array([event_index] * len(phase_type))

        # r = np.linalg.norm(event_loc[:, :2] - station_loc[:, :2], axis=-1, keepdims=False)
        # z = event_loc[:, 2] - station_loc[:, 2]
        x = events[event_index, 0] - stations[station_index, 0]
        y = events[event_index, 1] - stations[station_index, 1]
        z = events[event_index, 2] - stations[station_index, 2]
        r = np.sqrt(x**2 + y**2)

        rgrid0 = eikonal["rgrid"][0]
        zgrid0 = eikonal["zgrid"][0]
        nr = eikonal["nr"]
        nz = eikonal["nz"]
        h = eikonal["h"]

        if isinstance(phase_type, list):
            phase_type = np.array(phase_type)
        # if isinstance(station_index, list):
        #     station_index = np.array(station_index)

        tt = np.zeros(len(phase_type), dtype=np.float32)

        if isinstance(phase_type[0], str):
            p_index = phase_type == "P"
            s_index = phase_type == "S"
        elif isinstance(phase_type[0].item(), int):
            p_index = phase_type == 0
            s_index = phase_type == 1
        else:
            raise ValueError("phase_type must be either P/S or 0/1")

        if len(tt[p_index]) > 0:
            tt[p_index] = _interp(eikonal["up"], r[p_index], z[p_index], rgrid0, zgrid0, nr, nz, h)
        if len(tt[s_index]) > 0:
            tt[s_index] = _interp(eikonal["us"], r[s_index], z[s_index], rgrid0, zgrid0, nr, nz, h)

    return tt


def calc_traveltime(event_locs, station_locs, phase_type, eikonal):
    """
    event_locs: (num_event, 3) array of event locations
    station_locs: (num_station, 3) array of station locations
    phase_type: (num_event,) array of phase type
    eikonal: dictionary of eikonal solver
    """

    if eikonal is None:
        v = np.array([vel[x] for x in phase_type])
        tt = np.linalg.norm(event_locs - station_locs, axis=-1, keepdims=False) / v
    else:
        x = event_locs[:, 0] - station_locs[:, 0]
        y = event_locs[:, 1] - station_locs[:, 1]
        z = event_locs[:, 2] - station_locs[:, 2]
        r = np.sqrt(x**2 + y**2)

        rgrid0 = eikonal["rgrid"][0]
        zgrid0 = eikonal["zgrid"][0]
        nr = eikonal["nr"]
        nz = eikonal["nz"]
        h = eikonal["h"]

        if isinstance(phase_type, list):
            phase_type = np.array(phase_type)

        tt = np.zeros(len(phase_type), dtype=np.float32)
        if isinstance(phase_type[0], str):
            p_index = phase_type == "P"
            s_index = phase_type == "S"
        elif isinstance(phase_type[0].item(), int):
            p_index = phase_type == 0
            s_index = phase_type == 1
        else:
            raise ValueError("phase_type must be either P/S or 0/1")

        if len(tt[p_index]) > 0:
            tt[p_index] = _interp(eikonal["up"], r[p_index], z[p_index], rgrid0, zgrid0, nr, nz, h)
        if len(tt[s_index]) > 0:
            tt[s_index] = _interp(eikonal["us"], r[s_index], z[s_index], rgrid0, zgrid0, nr, nz, h)

    return tt


# def grad_traveltime(event_loc, station_loc, phase_type, eikonal):
def grad_traveltime(event_index, station_index, phase_type, events, stations, eikonal):

    if isinstance(event_index, int):
        event_index = np.array([event_index] * len(phase_type))

    # r = np.linalg.norm(event_loc[:, :2] - station_loc[:, :2], axis=-1, keepdims=False)
    # z = event_loc[:, 2] - station_loc[:, 2]
    x = events[event_index, 0] - stations[station_index, 0]
    y = events[event_index, 1] - stations[station_index, 1]
    z = events[event_index, 2] - stations[station_index, 2]
    r = np.sqrt(x**2 + y**2)

    rgrid0 = eikonal["rgrid"][0]
    zgrid0 = eikonal["zgrid"][0]
    nr = eikonal["nr"]
    nz = eikonal["nz"]
    h = eikonal["h"]

    if isinstance(phase_type, list):
        phase_type = np.array(phase_type)
    # if isinstance(station_index, list):
    #     station_index = np.array(station_index)

    dt_dr = np.zeros(len(phase_type))
    dt_dz = np.zeros(len(phase_type))

    # p_index = phase_type == "p"
    # s_index = phase_type == "s"
    p_index = phase_type == 0
    s_index = phase_type == 1

    dt_dr[p_index] = _interp(eikonal["grad_up"][0], r[p_index], z[p_index], rgrid0, zgrid0, nr, nz, h)
    dt_dr[s_index] = _interp(eikonal["grad_us"][0], r[s_index], z[s_index], rgrid0, zgrid0, nr, nz, h)
    dt_dz[p_index] = _interp(eikonal["grad_up"][1], r[p_index], z[p_index], rgrid0, zgrid0, nr, nz, h)
    dt_dz[s_index] = _interp(eikonal["grad_us"][1], r[s_index], z[s_index], rgrid0, zgrid0, nr, nz, h)

    # dr_dxy = (event_loc[:, :2] - station_loc[:, :2]) / (r[:, np.newaxis] + 1e-6)
    # dt_dxy = dt_dr[:, np.newaxis] * dr_dxy
    # grad = np.column_stack((dt_dxy, dt_dz[:, np.newaxis]))
    dt_dx = dt_dr * x / (r + 1e-6)
    dt_dy = dt_dr * y / (r + 1e-6)

    grad = np.column_stack((dt_dx, dt_dy, dt_dz))

    return grad


def init_eikonal2d(config):

    rlim = [
        0,
        np.sqrt(
            (config["xlim_km"][1] - config["xlim_km"][0]) ** 2 + (config["ylim_km"][1] - config["ylim_km"][0]) ** 2
        ),
    ]
    zlim = config["zlim_km"]
    h = config["h"]

    rgrid = np.arange(rlim[0], rlim[1] + h, h)
    zgrid = np.arange(zlim[0], zlim[1] + h, h)
    nr = len(rgrid)
    nz = len(zgrid)

    vel = config["vel"]
    zz, vp, vs = np.array(vel["Z"]), np.array(vel["P"]), np.array(vel["S"])

    # ##############################################
    # ## make the velocity staircase not linear
    # zz_grid = zz[1:] - h
    # vp_grid = vp[:-1]
    # vs_grid = vs[:-1]
    # zz = np.concatenate([zz, zz_grid])
    # vp = np.concatenate([vp, vp_grid])
    # vs = np.concatenate([vs, vs_grid])
    # idx = np.argsort(zz)
    # zz = zz[idx]
    # vp = vp[idx]
    # vs = vs[idx]
    # ##############################################

    vp1d = np.interp(zgrid, zz, vp)
    vs1d = np.interp(zgrid, zz, vs)
    vp = np.tile(vp1d, (nr, 1))
    vs = np.tile(vs1d, (nr, 1))
    # ir0 = np.floor(config["source_loc"][0] / h).astype(np.int64)
    # iz0 = np.floor(config["source_loc"][1] / h).astype(np.int64)
    ir0 = np.round(0 - rlim[0] / h).astype(np.int64)
    iz0 = np.round(0 - zlim[0] / h).astype(np.int64)
    up = 1000 * np.ones((nr, nz))
    # up[0, 0] = 0.0
    up[ir0, iz0] = 0.0

    up = eikonal_solve(up, vp, h)
    grad_up = np.gradient(up, h, edge_order=2)
    up = up.ravel()
    grad_up = [x.ravel() for x in grad_up]

    us = 1000 * np.ones((nr, nz))
    # us[0, 0] = 0.0
    us[ir0, iz0] = 0.0

    us = eikonal_solve(us, vs, h)
    grad_us = np.gradient(us, h, edge_order=2)
    us = us.ravel()
    grad_us = [x.ravel() for x in grad_us]

    config.update(
        {
            "up": up,
            "us": us,
            "grad_up": grad_up,
            "grad_us": grad_us,
            "rgrid": rgrid,
            "zgrid": zgrid,
            "nr": nr,
            "nz": nz,
            "h": h,
        }
    )

    return config


if __name__ == "__main__":

    import os

    data_path = "data"
    if not os.path.exists(data_path):
        os.makedirs(data_path, exist_ok=True)

    nr = 21
    nz = 21
    vel = {"p": 6.0, "s": 6.0 / 1.73}
    vp = np.ones((nr, nz)) * vel["p"]
    vs = np.ones((nr, nz)) * vel["s"]
    h = 1.0

    up = 1000 * np.ones((nr, nz))
    # up[nr//2, nz//2] = 0.0
    up[0, 0] = 0.0

    up = eikonal_solve(up, vp, h)
    grad_up = np.gradient(up, h, edge_order=2)
    up = up.ravel()
    grad_up = [x.ravel() for x in grad_up]

    us = 1000 * np.ones((nr, nz))
    # us[nr//2, nz//2] = 0.0
    us[0, 0] = 0.0

    us = eikonal_solve(us, vs, h)
    grad_us = np.gradient(us, h, edge_order=2)
    us = us.ravel()
    grad_us = [x.ravel() for x in grad_us]

    num_event = 10
    event_loc = np.random.rand(num_event, 3) * np.array([nr * h / np.sqrt(2), nr * h / np.sqrt(2), nz * h])
    event_index = np.arange(num_event)
    print(f"{event_loc = }")
    # event_loc = np.round(event_loc, 0)
    # station_loc = np.random.rand(1, 3) * np.array([nr*h/np.sqrt(2), nr*h/np.sqrt(2), 0])
    station_loc = np.array([0, 0, 0])
    print(f"{station_loc = }")
    station_loc = np.tile(station_loc, (num_event, 1))
    station_index = [0] * num_event
    # phase_type = np.random.choice(["p", "s"], num_event, replace=True)
    # print(f"{list(phase_type) = }")
    phase_type = np.array(["p"] * (num_event // 2) + ["s"] * (num_event - num_event // 2))
    v = np.array([vel[x] for x in phase_type])
    t = np.linalg.norm(event_loc - station_loc, axis=-1, keepdims=False) / v
    grad_t = (
        (event_loc - station_loc) / np.linalg.norm(event_loc - station_loc, axis=-1, keepdims=True) / v[:, np.newaxis]
    )
    print(f"True traveltime: {t = }")
    print(f"True grad traveltime: {grad_t = }")

    # tp = np.linalg.norm(event_loc - station_loc, axis=-1, keepdims=False) / vel["p"]
    # print(f"{tp = }")
    # ts = np.linalg.norm(event_loc - station_loc, axis=-1, keepdims=False) / vel["s"]
    # print(f"{ts = }")

    config = {
        "up": up,
        "us": us,
        "grad_up": grad_up,
        "grad_us": grad_us,
        "rgrid": np.arange(nr) * h,
        "zgrid": np.arange(nz) * h,
        "nr": nr,
        "nz": nz,
        "h": h,
    }
    mapping_int = {"p": 0, "s": 1}
    phase_type = np.array([mapping_int[x] for x in phase_type])
    # t = traveltime(event_loc, station_loc, phase_type, config)
    # grad_t = grad_traveltime(event_loc, station_loc, phase_type, config)
    t = traveltime(event_index, station_index, phase_type, event_loc, station_loc, config)
    grad_t = grad_traveltime(event_index, station_index, phase_type, event_loc, station_loc, config)
    print(f"Computed traveltime: {t = }")
    print(f"Computed grad traveltime: {grad_t = }")

    up = up.reshape((nr, nz))
    plt.figure()
    plt.pcolormesh(up[:, :])
    plt.gca().invert_yaxis()
    plt.colorbar()
    plt.savefig(f"{data_path}/slice_tp_2d.png")

    us = us.reshape((nr, nz))
    plt.figure()
    plt.pcolormesh(us[:, :])
    plt.gca().invert_yaxis()
    plt.colorbar()
    plt.savefig(f"{data_path}/slice_ts_2d.png")

    grad_up = [x.reshape((nr, nz)) for x in grad_up]
    fig, ax = plt.subplots(1, 2, figsize=(10, 5))
    cax0 = ax[0].pcolormesh(grad_up[0][:, :])
    fig.colorbar(cax0, ax=ax[0])
    ax[0].invert_yaxis()
    ax[0].set_title("grad_tp_x")
    cax1 = ax[1].pcolormesh(grad_up[1][:, :])
    fig.colorbar(cax1, ax=ax[1])
    ax[1].invert_yaxis()
    ax[1].set_title("grad_tp_z")
    plt.savefig(f"{data_path}/slice_grad_tp_2d.png")

    grad_us = [x.reshape((nr, nz)) for x in grad_us]
    fig, ax = plt.subplots(1, 2, figsize=(10, 5))
    cax0 = ax[0].pcolormesh(grad_us[0][:, :])
    fig.colorbar(cax0, ax=ax[0])
    ax[0].invert_yaxis()
    ax[0].set_title("grad_ts_x")
    cax1 = ax[1].pcolormesh(grad_us[1][:, :])
    fig.colorbar(cax1, ax=ax[1])
    ax[1].invert_yaxis()
    ax[1].set_title("grad_ts_z")
    plt.savefig(f"{data_path}/slice_grad_ts_2d.png")



# %%
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Function

from .eikonal2d import _interp
import scipy.sparse


class CalcTravelTime(Function):
    @staticmethod
    def forward(r, z, timetable, timetable_grad_r, timetable_grad_z, rgrid0, zgrid0, nr, nz, h):
        tt = _interp(timetable, r.numpy(), z.numpy(), rgrid0, zgrid0, nr, nz, h)
        tt = torch.from_numpy(tt)
        return tt

    @staticmethod
    def setup_context(ctx, inputs, output):
        r, z, timetable, timetable_grad_r, timetable_grad_z, rgrid0, zgrid0, nr, nz, h = inputs
        ctx.save_for_backward(r, z)
        ctx.timetable = timetable
        ctx.timetable_grad_r = timetable_grad_r
        ctx.timetable_grad_z = timetable_grad_z
        ctx.rgrid0 = rgrid0
        ctx.zgrid0 = zgrid0
        ctx.nr = nr
        ctx.nz = nz
        ctx.h = h

    @staticmethod
    def backward(ctx, grad_output):
        timetable_grad_r = ctx.timetable_grad_r
        timetable_grad_z = ctx.timetable_grad_z
        rgrid0 = ctx.rgrid0
        zgrid0 = ctx.zgrid0
        nr = ctx.nr
        nz = ctx.nz
        h = ctx.h
        r, z = ctx.saved_tensors

        grad_r = _interp(timetable_grad_r, r.numpy(), z.numpy(), rgrid0, zgrid0, nr, nz, h)
        grad_z = _interp(timetable_grad_z, r.numpy(), z.numpy(), rgrid0, zgrid0, nr, nz, h)

        grad_r = torch.from_numpy(grad_r) * grad_output
        grad_z = torch.from_numpy(grad_z) * grad_output

        return grad_r, grad_z, None, None, None, None, None, None, None, None


# %%
class Clamp(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, min, max):
        return input.clamp(min=min, max=max)

    @staticmethod
    def backward(ctx, grad_output):
        return grad_output, None, None


def clamp(input, min, max):
    return Clamp.apply(input, min, max)


def interp2d(time_table, x, y, xgrid, ygrid, h):
    nx = len(xgrid)
    ny = len(ygrid)
    assert time_table.shape == (nx, ny)

    with torch.no_grad():
        ix0 = torch.floor((x - xgrid[0]) / h).clamp(0, nx - 2).long()
        iy0 = torch.floor((y - ygrid[0]) / h).clamp(0, ny - 2).long()
        ix1 = ix0 + 1
        iy1 = iy0 + 1

    # x = (torch.clamp(x, xgrid[0], xgrid[-1]) - xgrid[0]) / h
    # y = (torch.clamp(y, ygrid[0], ygrid[-1]) - ygrid[0]) / h

    x = (clamp(x, xgrid[0], xgrid[-1]) - xgrid[0]) / h
    y = (clamp(y, ygrid[0], ygrid[-1]) - ygrid[0]) / h

    ## https://en.wikipedia.org/wiki/Bilinear_interpolation

    Q00 = time_table[ix0, iy0]
    Q01 = time_table[ix0, iy1]
    Q10 = time_table[ix1, iy0]
    Q11 = time_table[ix1, iy1]

    t = (
        Q00 * (ix1 - x) * (iy1 - y)
        + Q10 * (x - ix0) * (iy1 - y)
        + Q01 * (ix1 - x) * (y - iy0)
        + Q11 * (x - ix0) * (y - iy0)
    )

    return t


# %%
class TravelTime(nn.Module):
    def __init__(
        self,
        num_event,
        num_station,
        station_loc,
        station_dt=None,
        event_loc=None,
        event_time=None,
        event_loc0=None,
        event_time0=None,
        velocity={"P": 6.0, "S": 6.0 / 1.73},
        eikonal=None,
        zlim=[0, 30],
        dtype=torch.float32,
        grad_type="auto",
    ):
        super().__init__()
        self.num_event = num_event
        self.event_loc0 = nn.Embedding(num_event, 3)
        self.event_time0 = nn.Embedding(num_event, 1)
        self.event_loc = nn.Embedding(num_event, 3)
        self.event_time = nn.Embedding(num_event, 1)
        self.station_loc = nn.Embedding(num_station, 3)
        self.station_dt = nn.Embedding(num_station, 1)  # same statioin term for P and S

        ## initialize event_loc0 and event_time0
        if event_loc0 is not None:
            event_loc0 = torch.tensor(event_loc0, dtype=dtype).contiguous()
        else:
            event_loc0 = torch.zeros(num_event, 3, dtype=dtype).contiguous()
        if event_time0 is not None:
            event_time0 = torch.tensor(event_time0, dtype=dtype).contiguous()
        else:
            event_time0 = torch.zeros(num_event, 1, dtype=dtype).contiguous()
        self.event_loc0.weight = torch.nn.Parameter(event_loc0, requires_grad=False)
        self.event_time0.weight = torch.nn.Parameter(event_time0, requires_grad=False)

        ## check initialization
        station_loc = torch.tensor(station_loc, dtype=dtype)
        if station_dt is not None:
            station_dt = torch.tensor(station_dt, dtype=dtype)
        else:
            station_dt = torch.zeros(num_station, 1, dtype=dtype)
        if event_loc is not None:
            event_loc = torch.tensor(event_loc, dtype=dtype).contiguous()
        else:
            event_loc = torch.zeros(num_event, 3, dtype=dtype).contiguous()
        if event_time is not None:
            event_time = torch.tensor(event_time, dtype=dtype).contiguous()
        else:
            event_time = torch.zeros(num_event, 1, dtype=dtype).contiguous()

        self.station_loc.weight = torch.nn.Parameter(station_loc, requires_grad=False)
        self.station_dt.weight = torch.nn.Parameter(station_dt, requires_grad=False)
        self.event_loc.weight = torch.nn.Parameter(event_loc, requires_grad=True)
        self.event_time.weight = torch.nn.Parameter(event_time, requires_grad=True)

        # self.velocity = [velocity["P"], velocity["S"]]
        self.velocity = velocity
        self.eikonal = eikonal
        self.zlim = zlim
        self.grad_type = grad_type
        if self.eikonal is not None:
            self.timetable_p = torch.tensor(
                np.reshape(self.eikonal["up"], (self.eikonal["nr"], self.eikonal["nz"])), dtype=dtype
            )
            self.timetable_s = torch.tensor(
                np.reshape(self.eikonal["us"], (self.eikonal["nr"], self.eikonal["nz"])), dtype=dtype
            )
            self.rgrid = self.eikonal["rgrid"]
            self.zgrid = self.eikonal["zgrid"]
            self.h = self.eikonal["h"]

    def calc_time(self, event_loc, station_loc, phase_type):

        if self.eikonal is None:
            dist = torch.linalg.norm(event_loc - station_loc, axis=-1, keepdim=True)
            tt = dist / self.velocity[phase_type]
            tt = tt.float()
        else:
            # r = torch.linalg.norm(event_loc[:, :2] - station_loc[:, :2], axis=-1, keepdims=False)  ## nb, 3
            x = event_loc[:, 0] - station_loc[:, 0]
            y = event_loc[:, 1] - station_loc[:, 1]
            z = event_loc[:, 2] - station_loc[:, 2]
            r = torch.sqrt(x**2 + y**2)

            timetable = self.eikonal["up"] if phase_type in [0, "P"] else self.eikonal["us"]
            timetable_grad = self.eikonal["grad_up"] if phase_type in [0, "P"] else self.eikonal["grad_us"]
            timetable_grad_r = timetable_grad[0]
            timetable_grad_z = timetable_grad[1]
            rgrid0 = self.eikonal["rgrid"][0]
            zgrid0 = self.eikonal["zgrid"][0]
            nr = self.eikonal["nr"]
            nz = self.eikonal["nz"]
            h = self.eikonal["h"]
            tt = CalcTravelTime.apply(r, z, timetable, timetable_grad_r, timetable_grad_z, rgrid0, zgrid0, nr, nz, h)

            # if phase_type in [0, "P"]:
            #     timetable = self.timetable_p
            # elif phase_type in [1, "S"]:
            #     timetable = self.timetable_s
            # else:
            #     raise ValueError("phase_type should be 0 or 1. for P and S, respectively.")

            # tt = interp2d(timetable, r, z, self.rgrid, self.zgrid, self.h)

            tt = tt.float().unsqueeze(-1)

        return tt

    def forward(
        self,
        station_index,
        event_index=None,
        phase_type=None,
        phase_time=None,
        phase_weight=None,
    ):
        loss = 0.0
        pred_time = torch.zeros(len(phase_type), dtype=torch.float32)
        resisudal = torch.zeros(len(phase_type), dtype=torch.float32)

        # for type in [0, 1]:  # phase_type: 0 for P, 1 for S
        for type in np.unique(phase_type):

            if len(phase_type[phase_type == type]) == 0:
                continue

            station_index_ = station_index[phase_type == type]  # (nb,)
            event_index_ = event_index[phase_type == type]  # (nb,)
            phase_weight_ = phase_weight[phase_type == type]  # (nb,)

            station_loc_ = self.station_loc(station_index_)  # (nb, 3)
            station_dt_ = self.station_dt(station_index_)  # (nb, 1)

            event_loc_ = self.event_loc(event_index_)  # (nb, 3)
            event_time_ = self.event_time(event_index_)  # (nb, 1)

            tt_ = self.calc_time(event_loc_, station_loc_, type)  # (nb, 1)

            t_ = event_time_ + tt_ + station_dt_  # (nb, 1)
            t_ = t_.squeeze(1)  # (nb, )

            pred_time[phase_type == type] = t_  # (nb, )

            if phase_time is not None:
                phase_time_ = phase_time[phase_type == type]
                resisudal[phase_type == type] = phase_time_ - t_
                loss += torch.sum(F.huber_loss(t_, phase_time_, reduction="none") * phase_weight_)
                # loss += torch.sum(F.l1_loss(t_, phase_time_, reduction="none") * phase_weight_)
                # loss += torch.sum(torch.abs(t_ - phase_time_) * phase_weight_)

        return {"phase_time": pred_time, "residual": resisudal, "loss": loss}


# %%
class TravelTimeDD(TravelTime):

    def __init__(
        self,
        num_event,
        num_station,
        station_loc,
        station_dt=None,
        event_loc=None,
        event_time=None,
        event_loc0=None,
        event_time0=None,
        velocity={"P": 6.0, "S": 6.0 / 1.73},
        eikonal=None,
        zlim=[0, 30],
        dtype=torch.float32,
        grad_type="auto",
    ):
        super().__init__(
            num_event,
            num_station,
            station_loc,
            station_dt=station_dt,
            event_loc=event_loc,
            event_time=event_time,
            event_loc0=event_loc0,
            event_time0=event_time0,
            velocity=velocity,
            eikonal=eikonal,
            zlim=zlim,
            dtype=dtype,
            grad_type=grad_type,
        )

        self.event_loc.weight.requires_grad = True
        self.event_time.weight.requires_grad = False
        # self.event_time.weight.requires_grad = True

    def calc_time(self, event_loc, station_loc, phase_type):

        if self.eikonal is None:
            dist = torch.linalg.norm(event_loc - station_loc, axis=-1, keepdim=True)
            tt = dist / self.velocity[phase_type]
            tt = tt.float()
        else:
            nb1, ne1, nc1 = event_loc.shape  # batch, event, xyz
            nb2, ne2, nc2 = station_loc.shape
            assert ne1 % ne2 == 0
            assert nb1 == nb2
            station_loc = torch.repeat_interleave(station_loc, ne1 // ne2, dim=1)
            event_loc = event_loc.view(nb1 * ne1, nc1)
            station_loc = station_loc.view(nb1 * ne1, nc2)

            # r = torch.linalg.norm(event_loc[:, :2] - station_loc[:, :2], axis=-1, keepdims=False)  ## nb, 2 (pair), 3
            x = event_loc[:, 0] - station_loc[:, 0]
            y = event_loc[:, 1] - station_loc[:, 1]
            z = event_loc[:, 2] - station_loc[:, 2]
            r = torch.sqrt(x**2 + y**2)

            timetable = self.eikonal["up"] if phase_type in [0, "P"] else self.eikonal["us"]
            timetable_grad = self.eikonal["grad_up"] if phase_type in [0, "P"] else self.eikonal["grad_us"]
            timetable_grad_r = timetable_grad[0]
            timetable_grad_z = timetable_grad[1]
            rgrid0 = self.eikonal["rgrid"][0]
            zgrid0 = self.eikonal["zgrid"][0]
            nr = self.eikonal["nr"]
            nz = self.eikonal["nz"]
            h = self.eikonal["h"]
            tt = CalcTravelTime.apply(r, z, timetable, timetable_grad_r, timetable_grad_z, rgrid0, zgrid0, nr, nz, h)

            # if phase_type in [0, "P"]:
            #     timetable = self.timetable_p
            # elif phase_type in [1, "S"]:
            #     timetable = self.timetable_s
            # else:
            #     raise ValueError("phase_type should be 0 or 1. for P and S, respectively.")

            # tt = interp2d(timetable, r, z, self.rgrid, self.zgrid, self.h)

            tt = tt.float().view(nb1, ne1, 1)

        return tt

    def forward(
        self,
        station_index,
        event_index=None,
        phase_type=None,
        phase_time=None,
        phase_weight=None,
    ):
        loss = 0.0
        pred_time = torch.zeros(len(phase_type), dtype=torch.float32)

        # for type in [0, 1]:  # phase_type: 0 for P, 1 for S
        for type in np.unique(phase_type):

            if len(phase_type[phase_type == type]) == 0:
                continue

            station_index_ = station_index[phase_type == type]  # (nb,)
            event_index_ = event_index[phase_type == type]  # (nb,)
            phase_weight_ = phase_weight[phase_type == type]  # (nb,)

            station_loc_ = self.station_loc(station_index_)  # (nb, 3)
            station_dt_ = self.station_dt(station_index_)  # (nb, 1)

            delta_event_loc_ = self.event_loc(event_index_)  # (nb, 2, 3)
            delta_event_time_ = self.event_time(event_index_)  # (nb, 2, 1)

            event_loc0_ = self.event_loc0(event_index_)  # (nb, 2, 3)
            event_time0_ = self.event_time0(event_index_)  # (nb, 2, 1)

            station_loc_ = station_loc_.unsqueeze(1)  # (nb, 1, 3)
            station_dt_ = station_dt_.unsqueeze(1)  # (nb, 1, 1)

            event_loc_ = event_loc0_ + delta_event_loc_
            event_time_ = event_time0_ + delta_event_time_

            tt_ = self.calc_time(event_loc_, station_loc_, type)  # (nb, 2)

            t_ = event_time_ + tt_ + station_dt_  # (nb, 2, 1)

            t_ = t_[:, 0] - t_[:, 1]  # (nb, 1)
            t_ = t_.squeeze(1)  # (nb, )

            pred_time[phase_type == type] = t_  # (nb, )

            if phase_time is not None:
                phase_time_ = phase_time[phase_type == type]
                loss += torch.sum(F.huber_loss(t_, phase_time_, reduction="none") * phase_weight_)
                # loss += torch.sum(F.l1_loss(t_, phase_time_, reduction="none") * phase_weight_)
                # loss += torch.sum(torch.abs(t_ - phase_time_) * phase_weight_)

        if loss == 0.0:
            return None
        return {"phase_time": pred_time, "loss": loss}


def hypodd(pairs, event_loc, event_time, station_loc, config):

    num_event = len(event_loc)
    num_pair = len(pairs)
    pair_index = np.arange(num_pair)
    idx = np.ones(num_pair, dtype=bool)
    event_index = pairs[["idx_eve1", "idx_eve2"]].values
    station_index = pairs["idx_sta"].values
    phase_type = pairs["phase_type"].values
    phase_time = pairs["phase_dtime"].values
    phase_weight = pairs["phase_score"].values

    rgrid0 = config["eikonal"]["rgrid"][0]
    zgrid0 = config["eikonal"]["zgrid"][0]
    nr = config["eikonal"]["nr"]
    nz = config["eikonal"]["nz"]
    h = config["eikonal"]["h"]

    max_std = 10.0
    min_std = 3.0
    max_iter = 10
    for iter in range(max_iter):

        std = max_std - (max_std - min_std) * iter / max_iter

        rows = []
        cols = []
        values = []
        d = np.zeros(num_pair)

        station_loc_ = station_loc[station_index]
        event_loc_ = event_loc[event_index]
        event_time_ = event_time[event_index]

        station_loc_ = station_loc_[:, np.newaxis, :]
        x = event_loc_[:, :, 0] - station_loc_[:, :, 0]
        y = event_loc_[:, :, 1] - station_loc_[:, :, 1]
        z = event_loc_[:, :, 2] - station_loc_[:, :, 2]
        r = np.sqrt(x**2 + y**2)

        tt = np.zeros([len(phase_time), 2])
        grad_r = np.zeros([len(phase_time), 2])
        grad_z = np.zeros([len(phase_time), 2])

        for ph in np.unique(phase_type):
            timetable = config["eikonal"]["up"] if ph in [0, "P"] else config["eikonal"]["us"]
            timetable_grad = config["eikonal"]["grad_up"] if ph in [0, "P"] else config["eikonal"]["grad_us"]
            timetable_grad_r = timetable_grad[0]
            timetable_grad_z = timetable_grad[1]

            ii = phase_type == ph
            tt[ii] = _interp(timetable, r[ii], z[ii], rgrid0, zgrid0, nr, nz, h)
            grad_r[ii] = _interp(timetable_grad_r, r[ii], z[ii], rgrid0, zgrid0, nr, nz, h)
            grad_z[ii] = _interp(timetable_grad_z, r[ii], z[ii], rgrid0, zgrid0, nr, nz, h)

        grad_x = grad_r * x / (r + 1e-6)
        grad_y = grad_r * y / (r + 1e-6)
        dt = (tt[:, 0] + event_time_[:, 0] - tt[:, 1] - event_time_[:, 1]) - phase_time

        idx = np.abs(dt) < np.std(dt[idx]) * std
        dt = dt[idx]
        grad_x = grad_x[idx]
        grad_y = grad_y[idx]
        grad_z = grad_z[idx]
        num_pair_ = len(dt)
        pair_index_ = np.arange(num_pair_)
        event_index_ = event_index[idx]
        w = phase_weight[idx]

        rows = np.tile(pair_index_, 8)
        cols = np.transpose(event_index_)[:, np.newaxis, :] * 4 + np.arange(4)[np.newaxis, :, np.newaxis]
        values = np.concatenate(
            [
                grad_x[:, 0] * w,
                grad_y[:, 0] * w,
                grad_z[:, 0] * w,
                np.ones_like(w) * w,
                -grad_x[:, 1] * w,
                -grad_y[:, 1] * w,
                -grad_z[:, 1] * w,
                -np.ones_like(w) * w,
            ]
        )

        print(f"Number of pairs: {num_pair_} ({num_pair_/num_pair*100:.2f}%)")

        rows = rows.reshape(-1)
        cols = cols.reshape(-1)
        values = values.reshape(-1)

        G = scipy.sparse.coo_matrix((values, (rows, cols)), shape=(num_pair_, num_event * 4))
        d = dt * w

        x, istop, itn, r1norm, r2norm, anorm, acond, arnorm, xnorm, _ = scipy.sparse.linalg.lsqr(
            G, d, damp=0.0, conlim=100
        )
        x = x.reshape(num_event, 4)
        event_loc = event_loc - x[:, :3]
        event_time = event_time - x[:, 3]

        print(
            f"LSQR: {istop = }, {itn = }, {r1norm = :.3f}, {r2norm = :.3f}, {anorm = :.3f}, {acond = :.3f}, {arnorm = :.3f}, {xnorm = :.3f}"
        )

    return event_loc, event_time


# %%
if __name__ == "__main__":

    # %%
    import json
    import os
    from datetime import datetime, timedelta

    import matplotlib.pyplot as plt
    import pandas as pd
    from eikonal2d import eikonal_solve
    from torch import optim

    ######################################## Create Synthetic Data #########################################
    np.random.seed(0)
    data_path = "data"
    if not os.path.exists(data_path):
        os.makedirs(data_path, exist_ok=True)
    nx = 10
    ny = 10
    nr = int(np.sqrt(nx**2 + ny**2))
    nz = 10
    h = 3.0
    eikonal_config = {"nr": nr, "nz": nz, "h": h}
    with open(f"{data_path}/config.json", "w") as f:
        json.dump(eikonal_config, f)
    xgrid = np.arange(0, nx) * h
    ygrid = np.arange(0, ny) * h
    rgrid = np.arange(0, nr) * h
    zgrid = np.arange(0, nz) * h
    eikonal_config.update({"rgrid": rgrid, "zgrid": zgrid})
    num_station = 10
    num_event = 50
    stations = []
    for i in range(num_station):
        x = np.random.uniform(xgrid[0], xgrid[-1])
        y = np.random.uniform(ygrid[0], ygrid[-1])
        z = np.random.uniform(zgrid[0], zgrid[0] + 3 * h)
        stations.append({"station_id": f"STA{i:02d}", "x_km": x, "y_km": y, "z_km": z, "dt_s": 0.0})
    stations = pd.DataFrame(stations)
    stations["station_index"] = stations.index
    stations.to_csv(f"{data_path}/stations.csv", index=False)
    events = []
    reference_time = pd.to_datetime("2021-01-01T00:00:00.000")
    for i in range(num_event):
        x = np.random.uniform(xgrid[0], xgrid[-1])
        y = np.random.uniform(ygrid[0], ygrid[-1])
        z = np.random.uniform(zgrid[0], zgrid[-1])
        t = i * 5
        events.append(
            {"event_id": i, "event_time": reference_time + pd.Timedelta(seconds=t), "x_km": x, "y_km": y, "z_km": z}
        )
    events = pd.DataFrame(events)
    events["event_index"] = events.index
    events["event_time"] = events["event_time"].apply(lambda x: x.isoformat(timespec="milliseconds"))
    events.to_csv(f"{data_path}/events.csv", index=False)
    vpvs_ratio = 1.73
    vp = np.ones((nr, nz)) * 6.0
    vs = vp / vpvs_ratio

    ## eikonal solver
    up = 1000 * np.ones((nr, nz), dtype=np.float64)
    us = 1000 * np.ones((nr, nz), dtype=np.float64)
    ir0 = np.around((0 - rgrid[0]) / h).astype(np.int64)
    iz0 = np.around((0 - zgrid[0]) / h).astype(np.int64)
    up[ir0, iz0] = 0.0
    us[ir0, iz0] = 0.0
    up = eikonal_solve(up, vp, h)
    us = eikonal_solve(us, vs, h)
    up = torch.tensor(up, dtype=torch.float64)
    us = torch.tensor(us, dtype=torch.float64)

    # %%
    picks = []
    for j, station in stations.iterrows():
        for i, event in events.iterrows():
            r = np.linalg.norm([event["x_km"] - station["x_km"], event["y_km"] - station["y_km"]])
            z = event["z_km"] - station["z_km"]
            r = torch.tensor(r, dtype=torch.float64)
            z = torch.tensor(z, dtype=torch.float64)
            if np.random.rand() < 0.5:
                tt = interp2d(up, r, z, rgrid, zgrid, h).item()
                picks.append(
                    {
                        "event_id": event["event_id"],
                        "station_id": station["station_id"],
                        "phase_type": "P",
                        "phase_time": pd.to_datetime(event["event_time"]) + pd.Timedelta(seconds=tt),
                        "phase_score": 1.0,
                        "travel_time": tt,
                    }
                )
            if np.random.rand() < 0.5:
                tt = interp2d(us, r, z, rgrid, zgrid, h).item()
                picks.append(
                    {
                        "event_id": event["event_id"],
                        "station_id": station["station_id"],
                        "phase_type": "S",
                        "phase_time": pd.to_datetime(event["event_time"]) + pd.Timedelta(seconds=tt),
                        "phase_score": 1.0,
                        "travel_time": tt,
                    }
                )
    picks = pd.DataFrame(picks)
    # use picks,  stations.index, events.index to set station_index and
    picks["phase_time"] = picks["phase_time"].apply(lambda x: x.isoformat(timespec="milliseconds"))
    picks["event_index"] = picks["event_id"].map(events.set_index("event_id")["event_index"])
    picks["station_index"] = picks["station_id"].map(stations.set_index("station_id")["station_index"])
    picks.to_csv(f"{data_path}/picks.csv", index=False)
    # %%
    fig, ax = plt.subplots(1, 2, figsize=(12, 5))
    im = ax[0].imshow(vp[:, :], cmap="viridis")
    fig.colorbar(im, ax=ax[0])
    ax[0].set_title("Vp")
    im = ax[1].imshow(vs[:, :], cmap="viridis")
    fig.colorbar(im, ax=ax[1])
    ax[1].set_title("Vs")
    plt.savefig(f"{data_path}/true2d_vp_vs.png")

    # %%
    fig, ax = plt.subplots(1, 1, squeeze=False, figsize=(5, 5))
    ax[0, 0].scatter(stations["x_km"], stations["y_km"], c=stations["z_km"], marker="^", label="Station")
    ax[0, 0].scatter(events["x_km"], events["y_km"], c=events["z_km"], marker=".", label="Event")
    ax[0, 0].set_xlabel("x (km)")
    ax[0, 0].set_ylabel("y (km)")
    ax[0, 0].legend()
    ax[0, 0].set_title("Station and Event Locations")
    plt.savefig(f"{data_path}/station_event_3d.png")
    # %%
    fig, ax = plt.subplots(3, 1, squeeze=False, figsize=(10, 15))
    picks = picks.merge(stations, on="station_id")
    mapping_color = lambda x: f"C{int(x)}"
    ax[0, 0].scatter(pd.to_datetime(picks["phase_time"]), picks["x_km"], c=picks["event_index"].apply(mapping_color))
    ax[0, 0].scatter(
        pd.to_datetime(events["event_time"]), events["x_km"], c=events["event_index"].apply(mapping_color), marker="x"
    )
    ax[0, 0].set_xlabel("Time (s)")
    ax[0, 0].set_ylabel("x (km)")
    ax[1, 0].scatter(pd.to_datetime(picks["phase_time"]), picks["y_km"], c=picks["event_index"].apply(mapping_color))
    ax[1, 0].scatter(
        pd.to_datetime(events["event_time"]), events["y_km"], c=events["event_index"].apply(mapping_color), marker="x"
    )
    ax[1, 0].set_xlabel("Time (s)")
    ax[1, 0].set_ylabel("y (km)")
    ax[2, 0].scatter(pd.to_datetime(picks["phase_time"]), picks["z_km"], c=picks["event_index"].apply(mapping_color))
    ax[2, 0].scatter(
        pd.to_datetime(events["event_time"]), events["z_km"], c=events["event_index"].apply(mapping_color), marker="x"
    )
    ax[2, 0].set_xlabel("Time (s)")
    ax[2, 0].set_ylabel("z (km)")
    plt.savefig(f"{data_path}/picks_3d.png")
    # %%
    ######################################### Load Synthetic Data #########################################
    data_path = "data"
    events = pd.read_csv(f"{data_path}/events.csv")
    stations = pd.read_csv(f"{data_path}/stations.csv")
    picks = pd.read_csv(f"{data_path}/picks.csv")
    picks = picks.merge(events[["event_index", "event_time"]], on="event_index")

    #### make the time values relative to event time in seconds
    picks["phase_time_origin"] = picks["phase_time"].copy()
    picks["phase_time"] = (
        pd.to_datetime(picks["phase_time"]) - pd.to_datetime(picks["event_time"])
    ).dt.total_seconds()  # relative to event time (arrival time)
    picks.drop(columns=["event_time"], inplace=True)
    events["event_time_origin"] = events["event_time"].copy()
    events["event_time"] = np.zeros(len(events))  # relative to event time
    ####

    with open(f"{data_path}/config.json", "r") as f:
        eikonal_config = json.load(f)
    events = events.sort_values("event_index").set_index("event_index")
    stations = stations.sort_values("station_index").set_index("station_index")
    num_event = len(events)
    num_station = len(stations)

    ## eikonal solver
    nr, nz, h = eikonal_config["nr"], eikonal_config["nz"], eikonal_config["h"]
    rgrid = np.arange(0, nr) * h
    zgrid = np.arange(0, nz) * h
    eikonal_config.update({"rgrid": rgrid, "zgrid": zgrid})
    vp = np.ones((nr, nz), dtype=np.float64) * 6.0
    vs = vp / 1.73
    up = 1000 * np.ones((nr, nz), dtype=np.float64)
    us = 1000 * np.ones((nr, nz), dtype=np.float64)
    ir0 = np.around((0 - rgrid[0]) / h).astype(np.int64)
    iz0 = np.around((0 - zgrid[0]) / h).astype(np.int64)
    up[ir0, iz0] = 0.0
    us[ir0, iz0] = 0.0
    up = eikonal_solve(up, vp, h)
    us = eikonal_solve(us, vs, h)
    eikonal_config.update({"up": up, "us": us})

    ## initial event location
    event_loc = (
        events[["x_km", "y_km", "z_km"]].values * 0.0
    )  # + stations[["x_km", "y_km", "z_km"]].values.mean(axis=0)
    event_time = events[["event_time"]].values * 0.0

    # %%
    traveltime = TravelTime(
        num_event,
        num_station,
        stations[["x_km", "y_km", "z_km"]].values,
        stations[["dt_s"]].values,
        event_loc0=event_loc,
        event_time0=event_time,
        eikonal=eikonal_config,
    )

    # %%
    output = traveltime(
        torch.tensor(picks["station_index"].values, dtype=torch.long),
        torch.tensor(picks["event_index"].values, dtype=torch.long),
        picks["phase_type"].values,
        torch.tensor(picks["phase_time"].values, dtype=torch.float32),
        torch.tensor(picks["phase_score"].values, dtype=torch.float32),
    )
    loss = output["loss"]

    ######################################### Optimize #########################################
    # %%
    print(
        "Optimizing parameters:\n"
        + "\n".join(
            [f"{name}: {param.size()}" for name, param in traveltime.named_parameters() if param.requires_grad]
        ),
    )
    params = [param for param in traveltime.parameters() if param.requires_grad]
    optimizer = optim.LBFGS(params=params, max_iter=1000, line_search_fn="strong_wolfe")
    print("Initial loss:", loss.item())

    picks_station_index = torch.tensor(picks["station_index"].values, dtype=torch.long)
    picks_event_index = torch.tensor(picks["event_index"].values, dtype=torch.long)
    picks_phase_type = picks["phase_type"].values
    picks_phase_time = torch.tensor(picks["phase_time"].values, dtype=torch.float32)
    picks_phase_score = torch.tensor(picks["phase_score"].values, dtype=torch.float32)

    def closure():

        optimizer.zero_grad()

        output = traveltime(
            picks_station_index,
            picks_event_index,
            picks_phase_type,
            picks_phase_time,
            picks_phase_score,
        )
        loss = output["loss"]
        loss.backward()

        with torch.no_grad():
            traveltime.event_loc.weight.data[:, 2].clamp_(min=3.0)

        return loss

    optimizer.step(closure)

    output = traveltime(
        picks_station_index,
        picks_event_index,
        picks_phase_type,
        picks_phase_time,
        picks_phase_score,
    )
    loss = output["loss"]
    print("Final loss:", loss.item())

    # %%
    event_loc_init = event_loc.copy()
    event_loc = traveltime.event_loc.weight.detach().numpy()

    fig, ax = plt.subplots(1, 3, squeeze=False, figsize=(15, 5))
    # ax[0, 0].plot(stations["x_km"], stations["y_km"], "^", label="Station")
    ax[0, 0].plot(events["x_km"], events["y_km"], ".", label="True Events")
    ax[0, 0].plot(event_loc_init[:, 0], event_loc_init[:, 1], "x", label="Initial Events")
    for i in range(len(event_loc_init)):
        ax[0, 0].plot(
            [events["x_km"].iloc[i], event_loc_init[i, 0]],
            [events["y_km"].iloc[i], event_loc_init[i, 1]],
            "k--",
            alpha=0.5,
        )
    ax[0, 0].plot(event_loc[:, 0], event_loc[:, 1], "x", label="Inverted Events")
    for i in range(len(event_loc)):
        ax[0, 0].plot(
            [events["x_km"].iloc[i], event_loc[i, 0]], [events["y_km"].iloc[i], event_loc[i, 1]], "r--", alpha=0.5
        )
    ax[0, 0].set_xlabel("x (km)")
    ax[0, 0].set_ylabel("y (km)")
    ax[0, 0].legend()
    # ax[0, 0].set_title("Station and Event Locations")

    ax[0, 1].plot(events["x_km"], events["z_km"], ".", label="True Events")
    ax[0, 1].plot(event_loc_init[:, 0], event_loc_init[:, 2], "x", label="Initial Events")
    for i in range(len(event_loc_init)):
        ax[0, 1].plot(
            [events["x_km"].iloc[i], event_loc_init[i, 0]],
            [events["z_km"].iloc[i], event_loc_init[i, 2]],
            "k--",
            alpha=0.5,
        )
    ax[0, 1].plot(event_loc[:, 0], event_loc[:, 2], "x", label="Inverted Events")
    for i in range(len(event_loc)):
        ax[0, 1].plot(
            [events["x_km"].iloc[i], event_loc[i, 0]], [events["z_km"].iloc[i], event_loc[i, 2]], "r--", alpha=0.5
        )
    ax[0, 1].set_xlabel("x (km)")
    ax[0, 1].set_ylabel("z (km)")
    ax[0, 1].legend()

    ax[0, 2].plot(events["y_km"], events["z_km"], ".", label="True Events")
    ax[0, 2].plot(event_loc_init[:, 1], event_loc_init[:, 2], "x", label="Initial Events")
    for i in range(len(event_loc_init)):
        ax[0, 2].plot(
            [events["y_km"].iloc[i], event_loc_init[i, 1]],
            [events["z_km"].iloc[i], event_loc_init[i, 2]],
            "k--",
            alpha=0.5,
        )
    ax[0, 2].plot(event_loc[:, 1], event_loc[:, 2], "x", label="Inverted Events")
    for i in range(len(event_loc)):
        ax[0, 2].plot(
            [events["y_km"].iloc[i], event_loc[i, 1]], [events["z_km"].iloc[i], event_loc[i, 2]], "r--", alpha=0.5
        )
    ax[0, 2].set_xlabel("y (km)")
    ax[0, 2].set_ylabel("z (km)")
    ax[0, 2].legend()

    plt.savefig(f"{data_path}/inverted_station_event.png")



# %%
import matplotlib
import numpy as np
import scipy
from _ransac import RANSACRegressor
from eikonal3d import eikonal_solve, grad_traveltime, traveltime
from sklearn.base import BaseEstimator

# from sklearn.linear_model import RANSACRegressor
from tqdm import tqdm

np.random.seed(0)
# matplotlib.use("Agg")


class ADLoc(BaseEstimator):
    def __init__(self, config, stations, num_event=1, events=None, eikonal=None):
        """
        events: [x, y, z, t]
        """
        xlim = config["xlim_km"]
        ylim = config["ylim_km"]
        zlim = config["zlim_km"]
        vel = config["vel"]
        self.config = config
        self.stations = stations

        self.vel = vel
        self.num_event = num_event

        if eikonal is not None:
            if isinstance(eikonal["up"], str):
                eikonal["up"] = np.memmap(
                    eikonal["up"],
                    dtype="float32",
                    mode="r",
                    shape=(len(stations), eikonal["nx"] * eikonal["ny"] * eikonal["nz"]),
                )
            if isinstance(eikonal["us"], str):
                eikonal["us"] = np.memmap(
                    eikonal["us"],
                    dtype="float32",
                    mode="r",
                    shape=(len(stations), eikonal["nx"] * eikonal["ny"] * eikonal["nz"]),
                )
            if isinstance(eikonal["grad_up"], str):
                eikonal["grad_up"] = np.memmap(
                    eikonal["grad_up"],
                    dtype="float32",
                    mode="r",
                    shape=(len(stations), 3, eikonal["nx"] * eikonal["ny"] * eikonal["nz"]),
                )
            if isinstance(eikonal["grad_us"], str):
                eikonal["grad_us"] = np.memmap(
                    eikonal["grad_us"],
                    dtype="float32",
                    mode="r",
                    shape=(len(stations), 3, eikonal["nx"] * eikonal["ny"] * eikonal["nz"]),
                )
        self.eikonal = eikonal

        if events is not None:
            assert events.shape == (num_event, 4)
            self.events = events
        else:
            self.events = np.array(
                [[(xlim[0] + xlim[1]) / 2, (ylim[0] + ylim[1]) / 2, (zlim[0] + zlim[1]) / 2, 0]] * num_event
            )

    @staticmethod
    def loss_grad(event, X, y, vel={0: 6.0, 1: 6.0 / 1.75}, stations=None, eikonal=None):
        """
        X: data_frame with columns ["timestamp", "x_km", "y_km", "z_km", "type"]
        """
        station_index = X[:, 0]
        phase_type = X[:, 1]

        if eikonal is None:
            v = np.array([vel[t] for t in phase_type])
            tt = np.linalg.norm(event[:3] - stations[station_index, :3], axis=-1) / v + event[3]
        else:
            tt = traveltime(0, station_index, phase_type, event[np.newaxis, :3], stations, eikonal) + event[3]
        loss = 0.5 * np.sum((tt - y) ** 2)

        J = np.ones((len(X), 4))
        if eikonal is None:
            J[:, :3] = (
                (event[:3] - stations[station_index, :3])
                / np.linalg.norm(event[:3] - stations[station_index, :3], axis=-1, keepdims=True)
                / v[:, np.newaxis]
            )
        else:
            grad = grad_traveltime(0, station_index, phase_type, event[np.newaxis, :3], stations, eikonal)
            J[:, :3] = grad

        J = np.sum((tt - y)[:, np.newaxis] * J, axis=0)
        return loss, J

    def fit(self, X, y=None, event_index=0):

        opt = scipy.optimize.minimize(
            self.loss_grad,
            x0=self.events[event_index],
            method="L-BFGS-B",
            jac=True,
            args=(X, y, self.vel, self.stations, self.eikonal),
            bounds=[
                (self.config["xlim_km"][0], self.config["xlim_km"][1]),
                (self.config["ylim_km"][0], self.config["ylim_km"][1]),
                (self.config["zlim_km"][0], self.config["zlim_km"][1]),
                (None, None),
            ],
        )

        self.events[event_index, :] = opt.x
        self.is_fitted_ = True

        return self

    def predict(self, X, event_index=0):
        """
        X: data_frame with columns ["timestamp", "x_km", "y_km", "z_km", "type"]
        """
        station_index = X[:, 0]
        type = X[:, 1]

        if self.eikonal is None:
            v = np.array([self.vel[t] for t in type])
            tt = (
                np.linalg.norm(self.events[event_index, :3] - self.stations[station_index, :3], axis=-1) / v
                + self.events[event_index, 3]
            )
        else:
            tt = (
                traveltime(event_index, station_index, type, self.events, self.stations, self.eikonal)
                + self.events[event_index, 3]
            )

        return tt

    def score(self, X, y=None, event_index=0):
        """
        X: data_frame with columns ["timestamp", "x_km", "y_km", "z_km", "type"]
        """
        if len(X) == 0:
            return 0
        tt = self.predict(X, event_index)
        R2 = 1 - np.sum((y - tt) ** 2) / np.sum((y - np.mean(y)) ** 2)
        print(f"{R2=}")
        return R2


def init_eikonal3d(config, stations):

    # nr = 21
    # nz = 21
    # vel = {"p": 6.0, "s": 6.0 / 1.73}

    xlim = config["xlim_km"]
    ylim = config["ylim_km"]
    zlim = config["zlim_km"]
    # vel = config["vel"]
    if "vel" in config:
        vel = config["vel"]
    else:
        vel = {"p": 6.0, "s": 6.0 / 1.73}
    if "h" in config:
        h = config["h"]
    else:
        h = 2.0
    xgrid = np.arange(xlim[0], xlim[1], h)
    ygrid = np.arange(ylim[0], ylim[1], h)
    zgrid = np.arange(zlim[0], zlim[1], h)
    nx = len(xgrid)
    ny = len(ygrid)
    nz = len(zgrid)

    print(f"{nx=}, {ny=}, {nz=}")
    print(f"{len(stations) = }")
    vp = np.ones((nx, ny, nz)) * vel["p"]
    vs = np.ones((nx, ny, nz)) * vel["s"]

    num_station = len(stations)
    if os.path.exists("up.dat") or os.path.exists("us.dat"):
        if input("Files up.dat/us.data already exists. Do you want to overwrite? (y/n): ") == "y":
            up = np.memmap("up.dat", dtype="float32", mode="w+", shape=(num_station, nx * ny * nz))
            us = np.memmap("us.dat", dtype="float32", mode="w+", shape=(num_station, nx * ny * nz))
            grad_up = np.memmap("grad_up.dat", dtype="float32", mode="w+", shape=(num_station, 3, nx * ny * nz))
            grad_us = np.memmap("grad_us.dat", dtype="float32", mode="w+", shape=(num_station, 3, nx * ny * nz))

            pbar = tqdm(total=num_station)
            for i, station in stations.iterrows():

                pbar.set_description(f"Station {station['station_id']}")

                up_ = 1000 * np.ones((nx, ny, nz))

                us_ = 1000 * np.ones((nx, ny, nz))
                # up_[sta_ix, sta_iy, sta_iz] = 0.0
                # us_[sta_ix, sta_iy, sta_iz] = 0.0
                up_ = init_u0(xlim, ylim, zlim, station, h, vp, up_)
                us_ = init_u0(xlim, ylim, zlim, station, h, vs, us_)

                up_ = eikonal_solve(up_, vp, h)
                grad_up_ = np.gradient(up_, h, edge_order=2)
                up_ = up_.ravel()
                grad_up_ = [x.ravel() for x in grad_up_]
                up[i, :] = up_
                grad_up[i, :, :] = grad_up_

                # us_ = 1000 * np.ones((nx, ny, nz))
                # us_[sta_ix, sta_iy, sta_iz] = 0.0

                us_ = eikonal_solve(us_, vs, h)
                grad_us_ = np.gradient(us_, h, edge_order=2)
                us_ = us_.ravel()
                grad_us_ = [x.ravel() for x in grad_us_]
                us[i, :] = us_
                grad_us[i, :, :] = grad_us_

                pbar.update(1)

    config = {
        "up": "up.dat",
        "us": "us.dat",
        "grad_up": "grad_up.dat",
        "grad_us": "grad_us.dat",
        "xgrid": xgrid,
        "ygrid": ygrid,
        "zgrid": zgrid,
        "nx": nx,
        "ny": ny,
        "nz": nz,
        "h": h,
    }

    return config


def init_u0(xlim, ylim, zlim, station, h, v, u0):

    ix = (station["x_km"] - xlim[0]) / h
    iy = (station["y_km"] - ylim[0]) / h
    iz = (station["z_km"] - zlim[0]) / h
    ix0, ix1 = np.floor(ix).astype(int), np.ceil(ix).astype(int)
    iy0, iy1 = np.floor(iy).astype(int), np.ceil(iy).astype(int)
    iz0, iz1 = np.floor(iz).astype(int), np.ceil(iz).astype(int)

    u0[ix0, iy0, iz0] = np.sqrt((ix0 - ix) ** 2 + (iy0 - iy) ** 2 + (iz0 - iz) ** 2) * h / v[ix0, iy0, iz0]
    u0[ix1, iy0, iz0] = np.sqrt((ix1 - ix) ** 2 + (iy0 - iy) ** 2 + (iz0 - iz) ** 2) * h / v[ix1, iy0, iz0]
    u0[ix0, iy1, iz0] = np.sqrt((ix0 - ix) ** 2 + (iy1 - iy) ** 2 + (iz0 - iz) ** 2) * h / v[ix0, iy1, iz0]
    u0[ix0, iy0, iz1] = np.sqrt((ix0 - ix) ** 2 + (iy0 - iy) ** 2 + (iz1 - iz) ** 2) * h / v[ix0, iy0, iz1]
    u0[ix1, iy1, iz1] = np.sqrt((ix1 - ix) ** 2 + (iy1 - iy) ** 2 + (iz1 - iz) ** 2) * h / v[ix1, iy1, iz1]
    u0[ix0, iy1, iz1] = np.sqrt((ix0 - ix) ** 2 + (iy1 - iy) ** 2 + (iz1 - iz) ** 2) * h / v[ix0, iy1, iz1]
    u0[ix1, iy0, iz1] = np.sqrt((ix1 - ix) ** 2 + (iy0 - iy) ** 2 + (iz1 - iz) ** 2) * h / v[ix1, iy0, iz1]
    u0[ix1, iy1, iz0] = np.sqrt((ix1 - ix) ** 2 + (iy1 - iy) ** 2 + (iz0 - iz) ** 2) * h / v[ix1, iy1, iz0]

    return u0


# %%
if __name__ == "__main__":

    # %%
    import json
    import os

    import matplotlib.pyplot as plt
    import pandas as pd

    # data_path = "../tests/results"
    data_path = "./results"
    events = pd.read_csv(os.path.join(data_path, "events.csv"), parse_dates=["time"])
    stations = pd.read_csv(os.path.join(data_path, "stations.csv"))
    picks = pd.read_csv(os.path.join(data_path, "picks.csv"), parse_dates=["phase_time"])
    with open(os.path.join(data_path, "config.json"), "r") as f:
        config = json.load(f)

    eikonal = init_eikonal3d(config, stations)
    # eikonal = None

    # %%
    stations["idx_sta"] = stations.index  # reindex in case the index does not start from 0 or is not continuous

    events["idx_eve"] = events.index  # reindex in case the index does not start from 0 or is not continuous

    picks = picks.merge(events[["event_index", "idx_eve"]], on="event_index")

    picks = picks.merge(stations[["station_id", "idx_sta"]], on="station_id")

    # # %%
    # for idx_sta, picks_station in picks.groupby("idx_sta"):
    #     station_loc = stations.loc[idx_sta]
    #     print(f"Station {station_loc['station_id']} at ({station_loc['x_km']}, {station_loc['y_km']})")
    #     for _, pick in picks_station.iterrows():
    #         event_loc = events.loc[pick["idx_eve"]]
    #         print(f"Event {event_loc['event_index']} at ({event_loc['x_km']}, {event_loc['y_km']})")
    #     raise

    # %%
    event_index = 1
    num_event = len(events)
    # for idx_eve, picks_event in picks.groupby("idx_eve"):
    picks_event = picks[picks["idx_eve"] == event_index]
    event_loc = events.loc[event_index]
    print(f"Event {event_loc['event_index']} at ({event_loc['x_km']}, {event_loc['y_km']})")
    for _, pick in picks_event.iterrows():
        station_loc = stations.loc[pick["idx_sta"]]
        print(f"Station {station_loc['station_id']} at ({station_loc['x_km']}, {station_loc['y_km']})")

    # %%
    plt.figure()
    # plt.scatter(picks_event["x_km"], picks_event["y_km"], c=picks_event["phase_time"])
    tmp = picks_event.merge(stations[["x_km", "y_km", "z_km", "station_id"]], on="station_id")
    tmp["dist_km"] = tmp[["x_km", "y_km", "z_km"]].apply(
        lambda x: np.linalg.norm(x - event_loc[["x_km", "y_km", "z_km"]]), axis=1
    )
    plt.scatter(tmp["x_km"], tmp["y_km"], c=tmp["phase_time"], cmap="viridis")
    plt.scatter(event_loc["x_km"], event_loc["y_km"], c="r")
    plt.colorbar()
    plt.show()

    plt.figure()
    colors = lambda x: "r" if x == "P" else "b"
    plt.scatter(tmp["phase_time"], tmp["dist_km"], c=tmp["phase_type"].apply(colors))
    plt.show()

    # %%

    # %%
    config["vel"] = {"P": 6.0, "S": 6.0 / 1.73}
    X = picks_event.merge(
        stations[["x_km", "y_km", "z_km", "station_id"]],
        on="station_id",
    )
    t0 = X["phase_time"].min()
    X.rename(columns={"phase_type": "type", "phase_time": "t_s"}, inplace=True)
    X["t_s"] = (X["t_s"] - t0).dt.total_seconds()
    X = X[["x_km", "y_km", "z_km", "t_s", "type", "idx_sta"]]
    mapping_int = {"P": 0, "S": 1}
    config["vel"] = {mapping_int[k]: v for k, v in config["vel"].items()}
    X["type"] = X["type"].apply(lambda x: mapping_int[x.upper()])

    # estimator = ADLoc(config_, event_loc=np.array([x0, y0, z0]), event_t=(origin_time - t0).total_seconds())
    estimator = ADLoc(config, stations=stations[["x_km", "y_km", "z_km"]].values, num_event=num_event, eikonal=eikonal)
    # tt = estimator.predict(X[["x_km", "y_km", "z_km", "type"]].values)
    # estimator.score(X[["x_km", "y_km", "z_km", "type"]].values, y=X["t_s"].values)
    tt = estimator.predict(X[["idx_sta", "type"]].values, event_index=event_index)
    estimator.score(X[["idx_sta", "type"]].values, y=X["t_s"].values, event_index=event_index)

    print(f"True event loc: {event_loc[['x_km', 'y_km', 'z_km']].values}")
    print(f"Init event loc: {estimator.events[:3]}")

    # %%
    plt.figure()
    X["dist_km"] = X[["x_km", "y_km", "z_km"]].apply(
        lambda x: np.linalg.norm(x - estimator.events[event_index, :3]), axis=1
    )
    colors = lambda x: "r" if x == 0 else "b"
    plt.scatter(X["t_s"], X["dist_km"], c=X["type"].apply(colors), marker="x")
    plt.scatter(tt, X["dist_km"], c=X["type"].apply(colors), marker="o")
    plt.show()

    # %%
    # estimator.fit(X[["x_km", "y_km", "z_km", "type"]].values, y=X["t_s"].values)
    # estimator.score(X[["x_km", "y_km", "z_km", "type"]].values, y=X["t_s"].values)
    # tt = estimator.predict(X[["x_km", "y_km", "z_km", "type"]].values)
    estimator.fit(X[["idx_sta", "type"]].values, y=X["t_s"].values, event_index=event_index)
    estimator.score(X[["idx_sta", "type"]].values, y=X["t_s"].values, event_index=event_index)
    tt = estimator.predict(X[["idx_sta", "type"]].values)
    print(f"True event loc: {event_loc[['x_km', 'y_km', 'z_km']].values}")
    print(f"Estimated event loc: {estimator.events[:3]}")

    # %%
    plt.figure()
    X["dist_km"] = X[["x_km", "y_km", "z_km"]].apply(
        lambda x: np.linalg.norm(x - estimator.events[event_index, :3]), axis=1
    )
    colors = lambda x: "r" if x == 0 else "b"
    plt.scatter(X["t_s"], X["dist_km"], c=X["type"].apply(colors), marker="x")
    plt.scatter(tt, X["dist_km"], c=X["type"].apply(colors), marker="o")
    plt.show()

    # %%
    # reg = RANSACRegressor(estimator=ADLoc(config_, event_loc=np.array([[x0, y0, 0]]), event_t=0), random_state=0, min_samples=10, residual_threshold=4.0).fit(X[["x_km", "y_km", "z_km", "vel"]].values, X["t_s"].values)
    reg = RANSACRegressor(
        estimator=ADLoc(
            config, stations=stations[["x_km", "y_km", "z_km"]].values, num_event=num_event, eikonal=eikonal
        ),
        random_state=0,
        min_samples=4,
        residual_threshold=1.0,
    )
    # reg.fit(X[["x_km", "y_km", "z_km", "type"]].values, X["t_s"].values)
    reg.fit(X[["idx_sta", "type"]].values, X["t_s"].values, event_index=event_index)
    # reg = RANSACRegressor(estimator=ADLoc(config_), random_state=0, min_samples=4, residual_threshold=1.0).fit(X[["x_km", "y_km", "z_km", "type"]].values, X["t_s"].values)
    mask = reg.inlier_mask_
    estimator = reg.estimator_
    # estimator.score(X[["x_km", "y_km", "z_km", "type"]].values, y=X["t_s"].values)
    # tt = estimator.predict(X[["x_km", "y_km", "z_km", "type"]].values)
    estimator.score(X[["idx_sta", "type"]].values, y=X["t_s"].values, event_index=event_index)
    tt = estimator.predict(X[["idx_sta", "type"]].values, event_index=event_index)
    print(f"True event loc: {event_loc[['x_km', 'y_km', 'z_km']].values}")
    print(f"Estimated event loc: {estimator.events[:3]}")

    # %%
    plt.figure()
    X["dist_km"] = X[["x_km", "y_km", "z_km"]].apply(
        lambda x: np.linalg.norm(x - estimator.events[event_index, :3]), axis=1
    )
    colors = lambda x: "r" if x == 0 else "b"
    plt.scatter(X["t_s"], X["dist_km"], c=X["type"].apply(colors), marker="x")
    plt.scatter(tt, X["dist_km"], c=X["type"].apply(colors), marker="o")
    plt.show()

    # %% add random picks
    # %%
    config["vel"] = {"P": 6.0, "S": 6.0 / 1.73}
    X = picks_event.merge(stations[["x_km", "y_km", "z_km", "station_id"]], on="station_id")
    t0 = X["phase_time"].min()
    X.rename(columns={"phase_type": "type", "phase_time": "t_s"}, inplace=True)
    X["t_s"] = (X["t_s"] - t0).dt.total_seconds()
    X = X[["x_km", "y_km", "z_km", "t_s", "type", "idx_sta"]]
    mapping_int = {"P": 0, "S": 1}
    config["vel"] = {mapping_int[k]: v for k, v in config["vel"].items()}
    X["type"] = X["type"].apply(lambda x: mapping_int[x.upper()])

    num_noise = len(X) // 2
    noise = pd.DataFrame(
        {
            "x_km": np.random.rand(num_noise) * (X["x_km"].max() - X["x_km"].min()) + X["x_km"].min(),
            "y_km": np.random.rand(num_noise) * (X["y_km"].max() - X["y_km"].min()) + X["y_km"].min(),
            "z_km": np.random.rand(num_noise) * (X["z_km"].max() - X["z_km"].min()) + X["z_km"].min(),
            "t_s": np.random.rand(num_noise) * (X["t_s"].max() - X["t_s"].min()) + X["t_s"].min(),
            "type": np.random.choice([0, 1], num_noise),
            "idx_sta": np.random.choice(X["idx_sta"], num_noise),
            "mask": [0] * num_noise,
        }
    )
    X["mask"] = 1
    X = pd.concat([X, noise], ignore_index=True)
    X["dist_km"] = X[["x_km", "y_km", "z_km"]].apply(
        lambda x: np.linalg.norm(x - event_loc[["x_km", "y_km", "z_km"]]), axis=1
    )
    plt.figure()
    colors = lambda x: "r" if x == 0 else "b"
    plt.scatter(X["t_s"], X["dist_km"], c=X["type"].apply(colors), marker="o")
    plt.scatter(X["t_s"][X["mask"] == 0], X["dist_km"][X["mask"] == 0], c="k", marker="o", alpha=0.6)
    plt.show()

    # %%

    # reg = RANSACRegressor(estimator=ADLoc(config_, event_loc=np.array([[x0, y0, 0]]), event_t=0), random_state=0, min_samples=10, residual_threshold=4.0).fit(X[["x_km", "y_km", "z_km", "vel"]].values, X["t_s"].values)
    reg = RANSACRegressor(
        estimator=ADLoc(
            config, stations=stations[["x_km", "y_km", "z_km"]].values, num_event=num_event, eikonal=eikonal
        ),
        random_state=0,
        min_samples=4,
        residual_threshold=1.0,
    )
    # reg.fit(X[["x_km", "y_km", "z_km", "type"]].values, X["t_s"].values)
    reg.fit(X[["idx_sta", "type"]].values, X["t_s"].values, event_index=event_index)
    # reg = RANSACRegressor(estimator=ADLoc(config_), random_state=0, min_samples=4, residual_threshold=1.0).fit(X[["x_km", "y_km", "z_km", "type"]].values, X["t_s"].values)
    mask = reg.inlier_mask_
    estimator = reg.estimator_
    # estimator.score(X[["x_km", "y_km", "z_km", "type"]].values, y=X["t_s"].values)
    # tt = estimator.predict(X[["x_km", "y_km", "z_km", "type"]].values)
    estimator.score(X[["idx_sta", "type"]].values, y=X["t_s"].values, event_index=event_index)
    tt = estimator.predict(X[["idx_sta", "type"]].values, event_index=event_index)
    print(f"True event loc: {event_loc[['x_km', 'y_km', 'z_km']].values}")
    print(f"Estimated event loc: {estimator.events[:3]}")
    # %%
    plt.figure()
    X["dist_km"] = X[["x_km", "y_km", "z_km"]].apply(
        lambda x: np.linalg.norm(x - event_loc[["x_km", "y_km", "z_km"]]), axis=1
    )
    colors = lambda x: "r" if x == 0 else "b"
    plt.scatter(X["t_s"], X["dist_km"], c=X["type"].apply(colors), marker="o")
    plt.scatter(tt, X["dist_km"], c=X["type"].apply(colors), marker="x", s=100)
    plt.scatter(X["t_s"][~mask], X["dist_km"][~mask], c="k", marker="o", alpha=0.6)
    plt.show()

# %%



# %%
import numpy as np
import torch
import torch.nn as nn
from eikonal2d import _interp
from torch.autograd import Function


class Clamp(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, min, max):
        return input.clamp(min=min, max=max)

    @staticmethod
    def backward(ctx, grad_output):
        return grad_output, None, None


def clamp(input, min, max):
    return Clamp.apply(input, min, max)


class CalcTravelTime(Function):
    @staticmethod
    def forward(r, z, timetable, timetable_grad_r, timetable_grad_z, rgrid, zgrid, nr, nz, h):
        tt = _interp(timetable, r.numpy(), z.numpy(), rgrid[0], zgrid[0], nr, nz, h)
        tt = torch.from_numpy(tt)
        return tt

    @staticmethod
    def setup_context(ctx, inputs, output):
        r, z, timetable, timetable_grad_r, timetable_grad_z, rgrid, zgrid, nr, nz, h = inputs
        ctx.save_for_backward(r, z)
        ctx.timetable = timetable
        ctx.timetable_grad_r = timetable_grad_r
        ctx.timetable_grad_z = timetable_grad_z
        ctx.rgrid = rgrid
        ctx.zgrid = zgrid
        ctx.nr = nr
        ctx.nz = nz
        ctx.h = h

    @staticmethod
    def backward(ctx, grad_output):
        timetable_grad_r = ctx.timetable_grad_r
        timetable_grad_z = ctx.timetable_grad_z
        rgrid = ctx.rgrid
        zgrid = ctx.zgrid
        nr = ctx.nr
        nz = ctx.nz
        h = ctx.h
        r, z = ctx.saved_tensors

        grad_r = _interp(timetable_grad_r, r.numpy(), z.numpy(), rgrid[0], zgrid[0], nr, nz, h)
        grad_z = _interp(timetable_grad_z, r.numpy(), z.numpy(), rgrid[0], zgrid[0], nr, nz, h)

        grad_r = torch.from_numpy(grad_r) * grad_output
        grad_z = torch.from_numpy(grad_z) * grad_output

        return grad_r, grad_z, None, None, None, None, None, None, None, None


class Test(nn.Module):
    def __init__(self, timetable, rgrid, zgrid, grad_type="auto", timetable_grad_r=None, timetable_grad_z=None):
        super().__init__()
        self.timetable = timetable
        self.rgrid = rgrid
        self.zgrid = zgrid
        self.nr = len(rgrid)
        self.nz = len(zgrid)
        self.h = rgrid[1] - rgrid[0]
        assert self.h == zgrid[1] - zgrid[0]
        self.grad_type = grad_type
        self.timetable_grad_r = timetable_grad_r
        self.timetable_grad_z = timetable_grad_z

    def interp2d(self, time_table, r, z):
        nr = len(self.rgrid)
        nz = len(self.zgrid)
        assert time_table.shape == (nr, nz)

        ir0 = torch.floor((r - self.rgrid[0]) / self.h).clamp(0, nr - 2).long()
        iz0 = torch.floor((z - self.zgrid[0]) / self.h).clamp(0, nz - 2).long()
        ir1 = ir0 + 1
        iz1 = iz0 + 1

        r = (clamp(r, self.rgrid[0], self.rgrid[-1]) - self.rgrid[0]) / self.h
        z = (clamp(z, self.zgrid[0], self.zgrid[-1]) - self.zgrid[0]) / self.h
        # r = (torch.clamp(r, self.rgrid[0], self.rgrid[-1]) - self.rgrid[0]) / self.h
        # z = (torch.clamp(z, self.zgrid[0], self.zgrid[-1]) - self.zgrid[0]) / self.h

        ## https://en.wikipedia.org/wiki/Bilinear_interpolation
        Q00 = time_table[ir0, iz0]
        Q01 = time_table[ir0, iz1]
        Q10 = time_table[ir1, iz0]
        Q11 = time_table[ir1, iz1]

        t = (
            Q00 * (ir1 - r) * (iz1 - z)
            + Q10 * (r - ir0) * (iz1 - z)
            + Q01 * (ir1 - r) * (z - iz0)
            + Q11 * (r - ir0) * (z - iz0)
        )

        return t

    def forward(self, r, z):

        if self.grad_type == "auto":
            print(self.grad_type)
            tt = self.interp2d(self.timetable, r, z)
        else:
            tt = CalcTravelTime.apply(
                r,
                z,
                self.timetable,
                self.timetable_grad_r,
                self.timetable_grad_z,
                self.rgrid,
                self.zgrid,
                self.nr,
                self.nz,
                self.h,
            )

        return tt


if __name__ == "__main__":
    import time

    import matplotlib.pyplot as plt

    starttime = time.time()
    rgrid0 = 0
    zgrid0 = 0
    nr0 = 20
    nz0 = 20
    h = 1
    rgrid = rgrid0 + h * np.arange(0, nr0)
    zgrid = zgrid0 + h * np.arange(0, nz0)
    r, z = np.meshgrid(rgrid, zgrid, indexing="ij")
    timetable = np.sqrt(r**2 + z**2)
    grad_r, grad_z = np.gradient(timetable, h, edge_order=2)
    timetable = torch.from_numpy(timetable)
    # timetable = timetable.flatten()
    # grad_r = grad_r.flatten()
    # grad_z = grad_z.flatten()

    nr = 1000
    nz = 1000
    r = torch.linspace(-2, 22, nr)
    z = torch.linspace(-2, 22, nz)
    r, z = torch.meshgrid(r, z, indexing="ij")
    r = r.flatten()
    z = z.flatten()

    test = Test(
        timetable,
        rgrid,
        zgrid,
        grad_type="auto",
        timetable_grad_r=grad_r,
        timetable_grad_z=grad_z,
    )
    r.requires_grad = True
    z.requires_grad = True
    tt = test(r, z)
    tt.backward(torch.ones_like(tt))

    endtime = time.time()
    print(f"Time elapsed: {endtime - starttime} seconds.")
    tt = tt.detach().numpy()

    fig, ax = plt.subplots(3, 2)
    im = ax[0, 0].imshow(tt.reshape(nr, nz))
    fig.colorbar(im, ax=ax[0, 0])
    im = ax[0, 1].imshow(timetable.reshape(nr0, nz0))
    fig.colorbar(im, ax=ax[0, 1])
    im = ax[1, 0].imshow(r.grad.reshape(nr, nz))
    fig.colorbar(im, ax=ax[1, 0])
    im = ax[1, 1].imshow(grad_r.reshape(nr0, nz0))
    fig.colorbar(im, ax=ax[1, 1])
    im = ax[2, 0].imshow(z.grad.reshape(nr, nz))
    fig.colorbar(im, ax=ax[2, 0])
    im = ax[2, 1].imshow(grad_z.reshape(nr0, nz0))
    fig.colorbar(im, ax=ax[2, 1])
    plt.show()



from setuptools import setup

setup(
    name="ADLoc",
    version="1.0.2",
    long_description="ADLoc",
    long_description_content_type="text/markdown",
    packages=["adloc"],
    install_requires=["numpy", "h5py", "matplotlib", "pandas", "tqdm", "pyproj", "numba", "scipy"],
)
